# -*- coding: utf-8 -*-
"""
Created on Fri Mar 13 14:16:40 2020

@author: D Buckle

*******************************************************************************
Collection of classes to perform calculations of plan complexity metrics using
an input of mlcDataEval as produced by 'PlanOperations.py'

Metrics
    Base class for all the other metric classes to be derived from. The same
    operations will probably be needed for all metrics and this will save on
    repeating code.

ModulationComplexity
    Has methods to calculate the Leaf Sequence Variability and Aperture Area
    Variability as defined in the publication ??????? This calculation assumes
    the VMAT definition i.e. the average of two control points to take into
    account the moving jaws/leaves in between CP. This should still work for
    step/shoot IMRT as these have two CP/segment where the jaw/MLC positions
    are identical and the MU on the 1st one is 0.


BeamComplexity - Has methods which calculate:
    - aperture area (AA)
    - aperture perimeter (AP)
    - aperture irregularity (AI)
    - beam area (BA)
    - beam irregularity (BI)
    - beam modulation (BM)
*******************************************************************************
"""
import math
from math import pi

from pymedphys._imports import numpy as np

from . import data as MLCdata
from .operations import getBeamType, getMLCdata


class jawPositionError(ValueError):
    pass


class Metrics:
    def __init__(self, **kwargs):
        # initialiser function to define the default values of 0 for all parameters
        # and to pre-calculate the relevant parameters for the calculation
        self.ds = kwargs.get("ds")
        result = getMLCdata(self.ds)
        self.mlcData = result[0]
        self.jawData = result[1]
        self.segWeightData = result[2]
        self.maxSep = result[3]
        self.maxSepSum = result[4]
        self.maxPositions = result[5]
        self.maxJawPos = result[6]
        self.leafThicknesses = MLCdata.TrueBeamThick
        self.leafIndices = MLCdata.TrueBeam

        #        self.demographics = getDemographics()

        if self.mlcData is not None:
            self.beamNames = list(self.mlcData.keys())


#            self.controlPoints = len(self.mlcData.get(self.beamNames[0]))
#            self.leafIndices = self.mlcData.get(self.beamNames[0])[0]


def _calcLSV(left, right, y1, y2, indices):
    leftMax = abs(np.nanmax(left) - np.nanmin(left))
    rightMax = abs(np.nanmax(right) - np.nanmin(right))
    lsvLeft = 0
    lsvRight = 0
    leafDiffSum = 0
    leaves = 0
    assert len(left) == len(right)
    for l in range(len(left) - 1):  # looping through each leaf
        if indices[l] > float(y1) or indices[l] < float(y2):
            continue
        leafDiff = abs(left[l] - left[l + 1])
        leafDiffSum_ = leftMax - leafDiff
        leafDiffSum = leafDiffSum + leafDiffSum_
        leaves = leaves + 1
    try:  # handling for divide by 0 errors i.e. this factor needs to = 1 if no modulation
        lsvLeft = leafDiffSum / (leaves * leftMax)
    except ZeroDivisionError:
        lsvLeft = 1

    leafDiffSum = 0
    leaves = 0
    for l in range(len(right) - 1):
        leafDiff = abs(right[l] - right[l + 1])
        if math.isnan(leafDiff):
            continue
        leafDiffSum_ = rightMax - leafDiff
        leafDiffSum = leafDiffSum + leafDiffSum_
        leaves = leaves + 1

    try:
        lsvRight = leafDiffSum / (leaves * rightMax)
    except ZeroDivisionError:
        lsvRight = 1

    lsvSegment = lsvLeft * lsvRight
    return lsvSegment


class ModulationComplexity(Metrics):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.calcLSV = _calcLSV
        print("Beginning Modulation Complexity Score calculation....")

    def __str__(self):
        return str("str function to be coded later")

    def calcAAV(self, left, right, beamName, y1, y2, indices):
        leafSepSum = 0
        assert len(left) == len(right)
        for l in range(0, 60):  # looping round all the leaves
            if indices[l] > float(y1) or indices[l] < float(y2):
                continue
            leafSep = abs(right[l] - left[l])
            if leafSep > 0.5:
                leafSepSum = leafSepSum + leafSep
        try:
            m = self.maxSepSum.get(beamName)
            aavSegment = leafSepSum / m
        except (ZeroDivisionError, RuntimeWarning):
            aavSegment = 1

        return aavSegment

    def calcMCSs(self, b):
        mcsBeam = 0
        controlPoints = len(self.mlcData.get(b))
        for c in range(0, controlPoints - 1):
            indices = self.mlcData.get(b)[c][:, 0]
            A = self.mlcData.get(b)[c][:, 1]
            B = self.mlcData.get(b)[c][:, 2]
            y1 = self.jawData.get(b)[c][3, 1]
            y2 = self.jawData.get(b)[c][2, 1]
            lsvSeg = self.calcLSV(A, B, y1, y2, indices)
            aavSeg = self.calcAAV(A, B, b, y1, y2, indices)

            w = self.segWeightData.get(b)[c]

            mcsSeg = lsvSeg * aavSeg * w
            mcsBeam = mcsBeam + mcsSeg
        return mcsBeam

    #            print(f"MCS for beam {b}: ",round(mcsBeam,3))

    def calcMCSv(self, b):
        controlPoints = len(self.mlcData.get(b))
        mcsBeam = 0
        for c in range(0, controlPoints - 1):
            indices = self.mlcData.get(b)[c][:, 0]
            A = self.mlcData.get(b)[c][:, 1]
            B = self.mlcData.get(b)[c][:, 2]
            y1 = self.jawData.get(b)[c][3, 1]
            y2 = self.jawData.get(b)[c][2, 1]
            lsvSeg = self.calcLSV(A, B, y1, y2, indices)
            aavSeg = self.calcAAV(A, B, b, y1, y2, indices)

            indices = self.mlcData.get(b)[c + 1][:, 0]
            A = self.mlcData.get(b)[c + 1][:, 1]
            B = self.mlcData.get(b)[c + 1][:, 2]
            lsvSeg_ = self.calcLSV(A, B, y1, y2, indices)
            aavSeg_ = self.calcAAV(A, B, b, y1, y2, indices)

            l = (lsvSeg + lsvSeg_) / 2
            a = (aavSeg + aavSeg_) / 2
            w = self.segWeightData.get(b)[c]

            mcsSeg = l * a * w
            mcsBeam = mcsBeam + mcsSeg
        return mcsBeam

    def calcMCS(self):
        #        self.lsvPlan = {}
        #        self.aavPlan = {}
        beamNumber = 0
        mcsBeam = {}
        beamTypes = {}
        for b in self.beamNames:
            beamType = getBeamType(self.ds, beamNumber)
            beamNumber = beamNumber + 1
            #            print("Printing the beam type: ",beamType)
            if beamType == "STATIC":
                mcs = self.calcMCSs(b)
            if beamType == "DYNAMIC":
                mcs = self.calcMCSv(b)
            mcsBeam[b] = mcs
            beamTypes[b] = beamType
        print("Done.")
        return (mcsBeam, beamTypes)


#            print(f"MCS for beam {b}: ",round(mcsBeam,3))


def _calcApertureArea(indices, A, B, y1, y2):
    AASeg = 0
    for i in range(0, 60):
        if indices[i] > float(y1) or indices[i] < float(y2):
            continue
        if abs(A[i] - B[i]) < 1:
            continue
        AA_leafPair = abs(A[i] - B[i]) * MLCdata.TrueBeamThick[i]
        AA_leafPair = AA_leafPair / 100  # converting to cm
        AASeg = AASeg + AA_leafPair
    return AASeg


def _calcAperturePerim(indices, A, B, y1, y2):
    edgesSeg = 0
    endsSeg = 0
    for i in range(0, 60):
        # ignore leaf paris which are behind the jaws
        if indices[i] > float(y1) or indices[i] < float(y2):
            continue

        # check if leaf pair is open to get the contribution of the leaf
        # ends to the perimeter as twice the leaf thickness which is looked
        # up from the MLCData.py file
        if abs(A[i] - B[i]) > 5:
            endsSeg = endsSeg + 2 * MLCdata.TrueBeamThick[i]

        # work out the contribution of the leaf edges to the perimeter
        # Calculated by looking at the leaves above and below and checking
        # whether any part of either edge is contributing to the shape of the
        # segment. For the first and last leaf pairs there is no leaf above/below
        # so these need to be ignored.
        if i == 0:  # no leaf above so only checking one side
            if A[i + 1] < A[i]:
                edgesSeg = edgesSeg + abs(A[i + 1] - A[i])
            if B[i + 1] > B[i]:
                edgesSeg = edgesSeg + abs(B[i + 1] - B[i])
            continue
        if i == 60:  # no leaf below so only checking one side
            if A[i - 1] < A[i]:
                edgesSeg = edgesSeg + abs(A[i - 1] - A[i])
            if B[i - 1] > B[i]:
                edgesSeg = edgesSeg + abs(B[i - 1] - B[i])
            continue

        # checking both sides of each lead
        if A[i + 1] < A[i]:
            edgesSeg = edgesSeg + abs(A[i + 1] - A[i])
        if A[i - 1] > A[i]:
            edgesSeg = edgesSeg + abs(A[i - 1] - A[i])
        if B[i + 1] < B[i]:
            edgesSeg = edgesSeg + abs(B[i + 1] - B[i])
        if B[i + 1] > B[i]:
            edgesSeg = edgesSeg + abs(B[i + 1] - B[i])

    APSeg = (edgesSeg + endsSeg) / 10  # divide by 10 to convert to cm
    return APSeg


class BeamComplexity(Metrics):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.calcApertureArea = _calcApertureArea
        self.calcAperturePerim = _calcAperturePerim

        print("Beginning Beam Complexity calculation....")

    #        if self.jawData == None:
    #            print("Jaw positions not supplied. You must supply jaw position data.")
    #            raise jawPositionError(self.jawData)

    def __str__(self):
        return str("str function to be coded later")

    def calcApertureIrregularity(self, indices, A, B, y1, y2):
        AASeg = self.calcApertureArea(indices, A, B, y1, y2)
        APSeg = self.calcAperturePerim(indices, A, B, y1, y2)
        AISeg = (APSeg * APSeg) / (4 * pi * AASeg)
        return AASeg, APSeg, AISeg

    def calcBeamComplexity(self):
        beamArea = {}
        beamIrregularity = {}
        beamModulation = {}
        for b in self.beamNames:
            AABeam = 0
            AIBeam = 0
            beamMod = 0
            aMax = self.maxPositions.get(b)[:, 0]
            bMax = self.maxPositions.get(b)[:, 1]

            indices = self.mlcData.get(b)[0][:, 0]
            unionArea = self.calcApertureArea(indices, aMax, bMax, 200, -200)
            controlPoints = len(self.mlcData.get(b))
            for c in range(0, controlPoints):
                indices = self.mlcData.get(b)[c][:, 0]
                A = self.mlcData.get(b)[c][:, 1]
                B = self.mlcData.get(b)[c][:, 2]
                y1 = self.jawData.get(b)[c][3, 1]
                y2 = self.jawData.get(b)[c][2, 1]
                w = self.segWeightData.get(b)[c]

                AASeg = self.calcApertureArea(indices, A, B, y1, y2)
                AABeam = AABeam + (AASeg * w)

                AISeg = self.calcApertureIrregularity(indices, A, B, y1, y2)[2]
                AIBeam = AIBeam + (AISeg * w)

                beamMod = 1 - (AABeam / unionArea)

            beamArea[b] = AABeam
            beamIrregularity[b] = AIBeam
            beamModulation[b] = beamMod

        print("Done.")
        return (beamArea, beamIrregularity, beamModulation)



# Copyright (C) 2019 Matthew Jennings

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pymedphys._imports import numpy as np

from numpy import cos, radians, sin
from numpy.linalg import norm


def rotate_about_vector(coords_to_rotate, vector, theta, active=False):
    r"""Rotates a 3 x n vector of the form np.array((x, y, z)) about the axis specified
    by `vector`. Transforms can be active (alibi) or passive (alias). Default is
    passive.
    """
    unit_vector = vector / norm(vector)

    u_x = unit_vector[0]
    u_y = unit_vector[1]
    u_z = unit_vector[2]

    s = sin(radians(theta))
    c = cos(radians(theta))

    rotation_matrix = np.array(
        [
            [
                c + u_x * u_x * (1 - c),
                u_x * u_y * (1 - c) - u_z * s,
                u_x * u_z * (1 - c) + u_y * s,
            ],
            [
                u_y * u_x * (1 - c) + u_z * s,
                c + u_y * u_y * (1 - c),
                u_y * u_z * (1 - c) - u_x * s,
            ],
            [
                u_z * u_x * (1 - c) - u_y * s,
                u_z * u_y * (1 - c) + u_x * s,
                c + u_z * u_z * (1 - c),
            ],
        ]
    )

    # Rotation matrix above is active (unlike in other functions). Will manually
    # transpose to avoid confusion later...
    if not active:
        rotation_matrix = rotation_matrix.transpose()

    return rotation_matrix @ coords_to_rotate


def rotate_about_x(coords_to_rotate, psi, active=False):
    r"""Rotates a 3 x n vector of the form np.array((x, y, z)) about the x-axis.
    Transforms can be active (alibi) or passive (alias), but are passive by default.
    """
    s = sin(radians(psi))
    c = cos(radians(psi))

    x_rotation_matrix = np.array([[1, 0, 0], [0, c, s], [0, -s, c]])

    if active:
        x_rotation_matrix = x_rotation_matrix.transpose()

    return x_rotation_matrix @ coords_to_rotate


def rotate_about_y(coords_to_rotate, phi, active=False):
    r"""Rotates a 3 x n vector of the form np.array((x, y, z)) about the y-axis
    Transforms can be active (alibi) or passive (alias), but are passive by default.
    """
    s = sin(radians(phi))
    c = cos(radians(phi))

    y_rotation_matrix = np.array([[c, 0, -s], [0, 1, 0], [s, 0, c]])
    if active:
        y_rotation_matrix = y_rotation_matrix.transpose()

    return y_rotation_matrix @ coords_to_rotate


def rotate_about_z(coords_to_rotate, theta, active=False):
    r"""Rotates a 3 x n vector of the form np.array((x, y, z)) about the z-axis
    Transforms can be active (alibi) or passive (alias), but are passive by default.
    """
    s = sin(radians(theta))
    c = cos(radians(theta))

    z_rotation_matrix = np.array([[c, s, 0], [-s, c, 0], [0, 0, 1]])

    if active:
        z_rotation_matrix = z_rotation_matrix.transpose()

    return z_rotation_matrix @ coords_to_rotate


def translate(coords_to_translate, translation_vector, active=False):
    r"""Translates a 3 x Y array of the form np.array((x, y, z)) by a given
    displacement vector of the same form. Transforms can be active (alibi)
    or passive (alias), but are passive by default.
    """
    translation_dims = np.shape(coords_to_translate)

    for _ in translation_dims[1::]:
        translation_vector = np.expand_dims(translation_vector, axis=-1)

    if active:
        translation_vector = -translation_vector

    return coords_to_translate - translation_vector



# Copyright (C) 2018 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import os


def get_gantry_tolerance(index, file_hash, config):
    machine_name = index[file_hash]["logfile_header"]["machine"]
    machine_type = config["machine_map"][machine_name]["type"]
    gantry_tolerance = config["machine_types"][machine_type]["gantry_tolerance"]

    return gantry_tolerance


def get_data_directory(config):
    return config["linac_logfile_data_directory"]


def get_cache_filepaths(config):
    metersetmap_config = config["metersetmap"]

    comparison_storage_filepath = os.path.join(
        get_data_directory(config), metersetmap_config["comparisons_cache"]["primary"]
    )
    comparison_storage_scratch = os.path.join(
        get_data_directory(config), metersetmap_config["comparisons_cache"]["scratch"]
    )

    return comparison_storage_filepath, comparison_storage_scratch


def get_metersetmap_parameters(config):
    metersetmap_config = config["metersetmap"]
    grid_resolution = metersetmap_config["grid_resolution"]
    ram_fraction = metersetmap_config["ram_fraction"]

    return grid_resolution, ram_fraction


def get_index(config):
    index_filepath = os.path.join(get_data_directory(config), "index.json")
    with open(index_filepath) as json_data_file:
        index = json.load(json_data_file)

    return index


def get_centre(config, file_info):
    machine = file_info["logfile_header"]["machine"]
    centre = config["machine_map"][machine]["centre"]
    return centre


def get_sql_servers(config):
    centres = list(config["centres"].keys())

    sql_servers = {
        centre: config["centres"][centre]["ois_specific_data"]["sql_server"]
        for centre in centres
    }

    return sql_servers


def get_sql_servers_list(config):
    sql_servers = get_sql_servers(config)

    sql_servers_list = [value for _, value in sql_servers.items()]

    return sql_servers_list


def get_filepath(index, config, filehash):
    data_directory = get_data_directory(config)
    relative_path = index[filehash]["filepath"]
    filepath = os.path.abspath(os.path.join(data_directory, "indexed", relative_path))

    return filepath



from ._vendor.deprecated import deprecated as _deprecated


from .metersetmap import calculate as _calculate
from .metersetmap import display as _display
from .metersetmap import grid as _grid
from .metersetmap import WARNING_MESSAGE  # pylint: disable = unused-import


_calculate.__name__ = "pymedphys.mudensity.calculate"
calculate = _deprecated(
    reason="This has been replaced with `pymedphys.metersetmap.calculate`"
)(_calculate)

_display.__name__ = "pymedphys.mudensity.display"
display = _deprecated(
    reason="This has been replaced with `pymedphys.metersetmap.display`"
)(_display)

_grid.__name__ = "pymedphys.mudensity.grid"
grid = _deprecated(reason="This has been replaced with `pymedphys.metersetmap.grid`")(
    _grid
)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys._imports import numpy as np


def convert_IEC_angle_to_bipolar(angle):
    angle = np.copy(angle)
    if np.all(angle == 180):
        return angle

    angle[angle > 180] = angle[angle > 180] - 360

    is_180 = np.where(angle == 180)[0]
    not_180 = np.where(np.invert(angle == 180))[0]

    where_closest_left_leaning = np.argmin(
        np.abs(is_180[:, None] - not_180[None, :]), axis=1
    )
    where_closest_right_leaning = (
        len(not_180)
        - 1
        - np.argmin(np.abs(is_180[::-1, None] - not_180[None, ::-1]), axis=1)[::-1]
    )

    closest_left_leaning = not_180[where_closest_left_leaning]
    closest_right_leaning = not_180[where_closest_right_leaning]

    assert np.all(
        np.sign(angle[closest_left_leaning]) == np.sign(angle[closest_right_leaning])
    ), "Unable to automatically determine whether angle is 180 or -180"

    angle[is_180] = np.sign(angle[closest_left_leaning]) * angle[is_180]

    return angle



# Copyright (C) 2019 Paul King

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The following needs to be removed before leaving the experimental module
# pylint: skip-file


from pymedphys._imports import numpy as np

import matplotlib.image as mpimg
from PIL import Image


def read_narrow_png(file_name, step_size=0.1):
    """  Extract a an relative-density profilee from a narrow png file.

    Source file is a full color PNG that is sufficiently narrow that
    density uniform along its short dimension. The image density along
    its long dimension is reflective of a dose distribution. Requires
    Python PIL.

    Parameters
    ----------
    file_name : str
    step-size : float, optional
        Distance output increment in cm, defaults to 1 mm

    Returns
    -------
    array_like
        Image profile as a list of (distance, density) tuples, where density
        is an average color intensity value as represented in the source file.

    Raises
    ------
    ValueError
        If image is not narrow, i.e. aspect ratio <= 5
    AssertionError
        If step_size is too small, i.e. step_size <= 12.7 / dpi

    """

    image_file = Image.open(file_name)
    assert image_file.mode == "RGB"
    dpi_horiz, dpi_vert = image_file.info["dpi"]

    image_array = mpimg.imread(file_name)

    # DIMENSIONS TO AVG ACROSS DIFFERENT FOR HORIZ VS VERT IMG
    if image_array.shape[0] > 5 * image_array.shape[1]:  # VERT
        image_vector = np.average(image_array, axis=(1, 2))
        pixel_size_in_cm = 2.54 / dpi_vert
    elif image_array.shape[1] > 5 * image_array.shape[0]:  # HORIZ
        image_vector = np.average(image_array, axis=(0, 2))
        pixel_size_in_cm = 2.54 / dpi_horiz
    else:
        raise ValueError("The PNG file is not a narrow strip.")
    assert step_size > 5 * pixel_size_in_cm, "step size too small"

    if image_vector.shape[0] % 2 == 0:
        image_vector = image_vector[:-1]  # SO ZERO DISTANCE IS MID-PIXEL

    length_in_cm = image_vector.shape[0] * pixel_size_in_cm
    full_resolution_distances = np.arange(
        -length_in_cm / 2, length_in_cm / 2, pixel_size_in_cm
    )

    # TO MOVE FROM FILM RESOLUTION TO DESIRED PROFILE RESOLUTION
    num_pixels_to_avg_over = int(step_size / pixel_size_in_cm)
    sample_indices = np.arange(
        num_pixels_to_avg_over / 2,
        len(full_resolution_distances),
        num_pixels_to_avg_over,
    ).astype(int)
    downsampled_distances = list(full_resolution_distances[sample_indices])

    downsampled_density = []
    for idx in sample_indices:  # AVERAGE OVER THE SAMPLING WINDOW
        avg_density = np.average(
            image_vector[
                int(idx - num_pixels_to_avg_over / 2) : int(
                    idx + num_pixels_to_avg_over / 2
                )
            ]
        )
        downsampled_density.append(avg_density)

    zipped_profile = list(zip(downsampled_distances, downsampled_density))
    return zipped_profile



# Copyright (C) 2019 South Western Sydney Local Health District,
# University of New South Wales

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This work is derived from:
# https://github.com/AndrewWAlexander/Pinnacle-tar-DICOM
# which is released under the following license:

# Copyright (c) [2017] [Colleen Henschel, Andrew Alexander]

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
import os
import sys

from .image import convert_image
from .pinn_yaml import pinn_to_dict
from .pinnacle_image import PinnacleImage
from .pinnacle_plan import PinnaclePlan
from .rtdose import convert_dose
from .rtplan import convert_plan
from .rtstruct import convert_struct


class PinnacleExport:
    """Handle Pinnacle data to allow export of DICOM objects

    This class is used to export the various DICOM objects from Pinnacle
    data found within the given path.

    Parameters
    ----------
        path : str
            Path to raw Pinnacle data (directory containing 'Patient' file).
        logger : Logger
            Logger the tool will log to.
    """

    def __init__(self, path, logger=None):

        self._logger = logger  # Logger to use for all logging

        self._path = path  # Path to the root directory of the dataset
        # (containing the Patient file)

        self._patient_info = None  # The patient data read from
        self._plans = None  # Pinnacle plans for this path
        self._images = None  # Images found in image.info

        if not self._logger:
            self._logger = logging.getLogger(__name__)
            self._logger.setLevel(logging.DEBUG)

            if len(self._logger.handlers) == 0:
                ch = logging.StreamHandler(sys.stdout)
                formatter = logging.Formatter(
                    "%(asctime)s - %(levelname)s - %(message)s"
                )
                ch.setFormatter(formatter)
                ch.setLevel(logging.DEBUG)

                self._logger.addHandler(ch)

    @property
    def logger(self):
        """Gets the configured logger.

        Returns
        -------
        logger : Logger
            Logger configured.
        """
        return self._logger

    @property
    def patient_info(self):
        """Gets the patient info.

        Returns
        -------
        patient_info : dict
            Dictionary object containing the data found in the 'Patient'
            file within the raw Pinnacle data.
        """

        if not self._patient_info:
            path_patient = os.path.join(self._path, "Patient")
            self.logger.debug("Reading patient data from: %s", path_patient)
            self._patient_info = pinn_to_dict(path_patient)

            # Set the full patient name
            last_name = self._patient_info["LastName"]
            first_name = self._patient_info["FirstName"]
            middle_name = self._patient_info["MiddleName"]
            self._patient_info["FullName"] = f"{last_name}^{first_name}^{middle_name}^"

            # gets birthday string with numbers and dashes
            dobstr = self._patient_info["DateOfBirth"]
            if "-" in dobstr:
                dob_list = dobstr.split("-")
            elif "/" in dobstr:
                dob_list = dobstr.split("/")
            else:
                dob_list = dobstr.split(" ")

            dob = ""
            for num in dob_list:
                if len(num) == 1:
                    num = f"0{num}"
                dob = f"{dob}{num}"

            self._patient_info["DOB"] = dob

        return self._patient_info

    @property
    def plans(self):
        """Get the plans available.

        Returns
        -------
        plans : list
            List of PinnaclePlan's available within the Pinnacle raw data in the path
            configured.
        """

        # Read patient info to populate patients plans
        if not self._plans:

            self._plans = []
            for plan in self.patient_info["PlanList"]:
                path_plan = os.path.join(self._path, f"Plan_{plan['PlanID']}")
                self._plans.append(PinnaclePlan(self, path_plan, plan))

        return self._plans

    @property
    def images(self):
        """Get the images available.

        Returns
        -------
        images : list
            List of PinnacleImage's available within the Pinnacle raw data in the path
            configured.
        """

        # Read patient info to populate patients images
        if not self._images:
            self._images = []
            for image in self.patient_info["ImageSetList"]:
                pi = PinnacleImage(self, self._path, image)

                # Check that image info exists to ensure the image is really available
                if pi.image_info:
                    self._images.append(pi)

        return self._images

    @staticmethod
    def export_struct(plan, export_path="."):
        """Exports the RTSTRUCT DICOM modality.

        Parameters
        ----------
            plan : PinnaclePlan
                The plan for which structures should be exported.
            export_path : str, optional
                The file system path where the DICOM object should be
                exported to.
        """

        # Export Structures for plan
        convert_struct(plan, export_path)

    @staticmethod
    def export_dose(plan, export_path="."):
        """Exports the RTDOSE DICOM modality.

        Parameters
        ----------
            plan : PinnaclePlan
                The plan for which dose should be exported.
            export_path : str, optional
                The file system path where the DICOM object should be
                exported to.
        """

        # Export dose for plan
        convert_dose(plan, export_path)

    @staticmethod
    def export_plan(plan, export_path="."):
        """Exports the RTPLAN DICOM modality.

        Parameters
        ----------
            plan : PinnaclePlan
                The plan for which dose should be exported.
            export_path : str, optional
                The file system path where the DICOM object should be
                exported to.
        """

        # Export rtplan for plan
        convert_plan(plan, export_path)

    def export_image(self, image=None, series_uid="", export_path="."):
        """Exports an image from the Pinnacle data.

        Parameters
        ----------
            image : PinnacleImage, optional
                The image to export.
            series_uid : str, optional
                The SeriesInstanceUID of the image to export.
            export_path : str, optional
                The file system path where the DICOM objects should be
                exported to.
        """

        if not image and not series_uid:
            self.logger.error("No image to export")

        for im in self.images:
            im_info = im.image_info[0]
            im_suid = im_info["SeriesUID"]
            if len(series_uid) > 0 and im_suid == series_uid:
                convert_image(im, export_path)
                break

        if image:
            convert_image(image, export_path)

    def log_images(self):
        """Outputs all images found in the Pinnacle data to the log.
        """

        for i in self.images:
            image_header = i.image_header

            if not image_header:
                # If no ImageHeader is available then we can ignore this
                continue

            modality = "?"
            if "modality" in image_header.keys():
                modality = image_header["modality"]

            series_uid = "?"
            if "series_UID" in image_header.keys():
                series_uid = image_header["series_UID"]

            series_date_time = "?"
            if "SeriesDateTime" in image_header.keys():
                series_date_time = image_header["SeriesDateTime"]

            self.logger.info("%s: %s %s", modality, series_uid, series_date_time)

    def log_plan_names(self):
        """Outputs all plans found in the Pinnacle data to the log.
        """

        for p in self.plans:
            self.logger.info(p.plan_info["PlanName"])

    def log_trial_names(self):
        """Outputs all trials found in the Pinnacle data to the log.
        """

        for p in self.plans:
            self.logger.info("### %s ###", p.plan_info["PlanName"])
            for t in p.trials:
                self.logger.info("- %s", t["Name"])

    def log_trial_names_in_plan(self, plan):
        """Outputs all trials found within a plan to the log.

        Parameters
        ----------
            plan : PinnaclePlan
                The plan for which to log the trials.
        """

        self.logger.info("### %s ###", plan.plan_info["PlanName"])
        self.logger.info(plan.path)
        for trial in plan.trials:
            self.logger.info("- %s", trial["Name"])



# Copyright (C) 2020 South Western Sydney Local Health District,
# University of New South Wales

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This work is derived from:
# https://github.com/AndrewWAlexander/Pinnacle-tar-DICOM
# which is released under the following license:

# Copyright (c) [2017] [Colleen Henschel, Andrew Alexander]

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylint: disable = redefined-outer-name

import os
import subprocess
import tempfile
from pathlib import Path
from zipfile import ZipFile

import pytest

from pymedphys._data import download

working_path = tempfile.mkdtemp()
data_path = os.path.join(working_path, "data")


def get_online_data(filename):
    return download.get_file_within_data_zip("pinnacle_test_data.zip", filename)


@pytest.fixture(scope="session")
def data():

    zip_ref = ZipFile(get_online_data("pinnacle_16.0_test_data.zip"), "r")
    zip_ref.extractall(data_path)
    zip_ref.close()

    return Path(data_path)


@pytest.mark.slow
def test_pinnacle_cli_output(data):

    output_path = tempfile.mkdtemp()

    for pinn_dir in data.joinpath("Pt1").joinpath("Pinnacle").iterdir():

        command = "pymedphys experimental pinnacle export".split() + [
            "-o",
            output_path,
            "-m",
            "CT",
            "-m",
            "RTSTRUCT",
            "-m",
            "RTDOSE",
            "-m",
            "RTPLAN",
            "-t",
            "Trial_1",
            pinn_dir.as_posix(),
        ]

        subprocess.check_call(command)

    # Just check that the number of output files is what we expect
    # The output itself is checked in the other function tests
    assert len(os.listdir(output_path)) == 119


@pytest.mark.slow
def test_pinnacle_cli_list(data):

    for pinn_dir in data.joinpath("Pt1").joinpath("Pinnacle").iterdir():

        command = "pymedphys experimental pinnacle export".split() + [
            "-l",
            pinn_dir.as_posix(),
        ]

        cli_output = str(subprocess.check_output(command))
        cli_output_parts = cli_output.split("\\n")
        assert "Plans and Trials" in cli_output_parts[0]
        assert "Plan_0" in cli_output_parts[1]
        assert "Trial_1" in cli_output_parts[2]
        assert "Images" in cli_output_parts[3]
        assert (
            "CT: 1.2.826.0.1.3680043.9.7225.3631975141391052922211556418733774032"
            in cli_output_parts[4]
        )


@pytest.mark.slow
def test_pinnacle_cli_missing_trial(data):

    output_path = tempfile.mkdtemp()

    for pinn_dir in data.joinpath("Pt1").joinpath("Pinnacle").iterdir():

        command = "pymedphys experimental pinnacle export".split() + [
            "-o",
            output_path,
            "-t",
            "nonexistenttrial",
            pinn_dir.as_posix(),
        ]

        cli_output = str(subprocess.check_output(command))
        assert "No Trial: nonexistenttrial found in Plan" in cli_output



from .header import determine_header_length


def split_into_header_table(trf_contents):
    header_length = determine_header_length(trf_contents)

    trf_header_contents = trf_contents[0:header_length]
    trf_table_contents = trf_contents[header_length::]

    return trf_header_contents, trf_table_contents



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# The following needs to be removed before leaving the experimental module
# pylint: skip-file

import json
import os

import pytest
from fixtures import BASELINES_DIR, postscans, prescans

import matplotlib.pyplot as plt
import numpy as np

from pymedphys._experimental.film import (
    create_axes,
    get_aligned_image,
    shift_and_rotate,
)

CREATE_BASELINE = False

ALIGNMENT_BASELINES_FILEPATH = os.path.join(BASELINES_DIR, "pre_post_alignment.json")


def test_multi_channel_shift_and_rotate(
    prescans,
):  # pylint: disable=redefined-outer-name
    prescan = prescans[0]
    axes = create_axes(prescan)

    interpolated = shift_and_rotate(axes, axes, prescan, 0, 0, 0)

    assert np.allclose(interpolated, prescan)


def get_alignment(prescan, postscan, baseline=None):
    shifted_prescan, alignment = get_aligned_image(prescan, postscan)

    if baseline is None or not np.allclose(baseline, alignment, 0.01, 0.01):
        print(baseline)
        print(alignment)

        plt.figure()
        plt.imshow(postscan)

        plt.figure()
        plt.imshow(prescan)

        plt.figure()
        plt.imshow(shifted_prescan.astype(np.uint8))

        plt.show()

        if baseline is not None:
            raise AssertionError

    return alignment


def test_pre_and_post_align(
    prescans, postscans
):  # pylint: disable=redefined-outer-name
    keys = prescans.keys()
    assert keys == postscans.keys()

    if not CREATE_BASELINE:
        with open(ALIGNMENT_BASELINES_FILEPATH, "r") as a_file:
            baselines = json.load(a_file)
    else:
        baselines = {str(key): None for key in keys}

    results = {}

    # keys_to_use = keys
    keys_to_use = [0.0, 1000.0]

    for key in keys_to_use:
        results[key] = np.around(
            get_alignment(prescans[key], postscans[key], baseline=baselines[str(key)]),
            decimals=4,
        ).tolist()

    if CREATE_BASELINE:
        with open(ALIGNMENT_BASELINES_FILEPATH, "w") as a_file:
            json.dump(results, a_file)



# Copyright (C) 2020 Rafael Ayala

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Export csv data from QuickCheck measurements.
"""

from pymedphys._experimental.quickcheck import export_cli


def quickcheck_cli(subparsers):
    quickcheck_parser = subparsers.add_parser(
        "quickcheck",
        help="A toolbox to retrieve measurement data from PTW Quickcheck and write it to a csv file.",
    )
    quickcheck_subparsers = quickcheck_parser.add_subparsers(dest="quickcheck")

    export_quickcheck(quickcheck_subparsers)

    return quickcheck_parser


def export_quickcheck(quickcheck_subparsers):
    parser = quickcheck_subparsers.add_parser(
        "to-csv", help="Export Quickcheck data to .csv"
    )

    parser.add_argument(
        "ip", type=str, help="IP address or host name of Quickcheck device"
    )

    parser.add_argument("csv_path", type=str, help="file path for csv file.")

    parser.set_defaults(func=export_cli)



# Copyright (C) 2019 South Western Sydney Local Health District,
# University of New South Wales

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Export DICOM objects from raw Pinnacle data.
"""

from pymedphys._experimental.pinnacle import export_cli


def pinnacle_cli(subparsers):
    pinnacle_parser = subparsers.add_parser(
        "pinnacle", help="A toolbox to export raw Pinnacle data to DICOM."
    )
    pinnacle_subparsers = pinnacle_parser.add_subparsers(dest="pinnacle")

    export_pinnacle(pinnacle_subparsers)

    return pinnacle_parser


def export_pinnacle(pinnacle_subparsers):
    parser = pinnacle_subparsers.add_parser("export", help="Export a raw file to DICOM")

    parser.add_argument(
        "input_path",
        type=str,
        help=(
            "Root Patient directory of raw Pinnacle data (directory "
            "containing the 'Patient' file). Alternatively a TAR archive "
            "can be supplied."
        ),
    )

    parser.add_argument(
        "-o",
        "--output-directory",
        help=("Directory in which to generate DICOM objects."),
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help=("Flag to output debug information."),
    )

    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help=("List all plans and trials available."),
    )

    parser.add_argument(
        "-m",
        "--modality",
        action="append",
        default=[],
        help=("Modalities to export (CT exports the plans primary " "planning CT)."),
    )

    parser.add_argument(
        "-p",
        "--plan",
        help=(
            "The name of the plan to export (first plan will be "
            "exported by default)."
        ),
    )

    parser.add_argument(
        "-t",
        "--trial",
        help=(
            "The name of the trial to export (first trial will be "
            "exported by default)."
        ),
    )

    parser.add_argument(
        "-i",
        "--image",
        help=(
            "The UID of an image series you would like to export "
            "(or 'all' to export all images)."
        ),
    )

    parser.add_argument(
        "-u", "--uid-prefix", help=("Prefix to use for generated UIDs.")
    )

    parser.set_defaults(func=export_cli)



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import time

import streamlit as st

from pymedphys._streamlit.apps import anonymise_monaco as _anonymise_monaco
from pymedphys._streamlit.apps import dashboard as _dashboard
from pymedphys._streamlit.apps import electrons as _electrons
from pymedphys._streamlit.apps import metersetmap as _metersetmap
from pymedphys._streamlit.apps import pseudonymise as _pseudonymise
from pymedphys._streamlit.utilities import session

HERE = pathlib.Path(__file__).parent.resolve()
FAVICON = str(HERE.joinpath("pymedphys.png"))

APPLICATION_CATEGORIES = {
    "mature": {
        "title": "Mature",
        "description": """
            These are mature applications. They are in wide use, and
            they have a high level of automated test coverage.
        """,
    },
    "maturing": {
        "title": "Maturing",
        "description": """
            These are relatively new applications. They potentially
            only have limited use within the community, but the still
            adhere to high quality standards with a level of automated
            test coverage that can be expected for a mature application.
        """,
    },
    "raw": {
        "title": "Raw",
        "description": """
            These are relatively new applications. They possibly only
            have minimal use within the community, and they have at
            least some automated test coverage. It is likely that these
            applications and their respective configurations will still
            be changing as time goes on.
        """,
    },
    "beta": {
        "title": "Beta",
        "description": """
            These applications may not be in use at all within the
            community. They potentially may only have minimal automated
            test coverage.
        """,
    },
    "experimental": {
        "title": "Experimental",
        "description": """
            These applications may not be in use at all within the
            community and they may not have any automated test coverage.
            They may not even work.
        """,
    },
}


APPLICATION_OPTIONS = {
    "metersetmap": {
        "category": "raw",
        "label": "MetersetMap Comparison",
        "callable": _metersetmap.main,
    },
    "pseudonymise": {
        "category": "raw",
        "label": "DICOM Pseudonymisation",
        "callable": _pseudonymise.main,
    },
    "dashboard": {
        "category": "experimental",
        "label": "Clinical Dashboard",
        "callable": _dashboard.main,
    },
    "electrons": {
        "category": "experimental",
        "label": "Electron Insert Factor Modelling",
        "callable": _electrons.main,
    },
    "anonymise-monaco": {
        "category": "experimental",
        "label": "Anonymising Monaco Backend Files",
        "callable": _anonymise_monaco.main,
    },
}

st.set_page_config(page_title="PyMedPhys", page_icon=FAVICON)


def get_url_app():
    try:
        return st.experimental_get_query_params()["app"][0]
    except KeyError:
        return "index"


session_state = session.initialise_session_state(app=get_url_app())


def swap_app(app):
    st.experimental_set_query_params(app=app)
    session_state.app = app

    # Not sure why this is needed. The `set_query_params` doesn't
    # appear to work if a rerun is undergone immediately afterwards.
    time.sleep(0.01)
    st.experimental_rerun()


def index():
    st.write(
        """
        # Index of applications available

        The following applications are organised by category where each
        category is representative of the maturity of the tool.
        """
    )

    for category_key, category in APPLICATION_CATEGORIES.items():
        st.write(
            f"""
                ## {category["title"]}
                {category["description"]}
            """
        )

        st.write("---")

        applications_in_this_category = [
            item
            for item in APPLICATION_OPTIONS.items()
            if item[1]["category"] == category_key
        ]

        if not applications_in_this_category:
            st.write("> *No applications are currently in this category.*")

        for app_key, application in applications_in_this_category:
            if st.button(application["label"]):
                swap_app(app_key)

        st.write("---")


def main():
    if (
        session_state.app != "index"
        and not session_state.app in APPLICATION_OPTIONS.keys()
    ):
        swap_app("index")

    if session_state.app != "index":
        if st.sidebar.button("Return to Index"):
            swap_app("index")

        st.sidebar.write("---")

    if session_state.app == "index":
        application_function = index
    else:
        application_function = APPLICATION_OPTIONS[session_state.app]["callable"]

    application_function()


if __name__ == "__main__":
    main()



#############################START LICENSE##########################################
# Copyright (C) 2019 Pedro Martinez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#############################END LICENSE##########################################


###########################################################################################
#
#   Script name: qc-graticule
#
#   Description: Tool for calculating graticule centre at different gantry angles.
#
#   Example usage: python qc-graticule "/folder/"
#
#   The folder can contain:
#   1/2 image(s) at g=0
#   1/2 image(s) at g=90
#   1/2 image(s) at g=180
#   1/2 image(s) at g=270
#
#   Author: Pedro Martinez
#   pedro.enrique.83@gmail.com
#   5877000722
#   Date:2020-05-12
#
###########################################################################################

import argparse
import os
import sys
from datetime import datetime
from operator import itemgetter
from sys import platform

from tqdm import tqdm

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
from PIL import Image
from skimage.feature import blob_log

import pydicom

from .utils import utils as u


def running_mean(x, N):
    out = np.zeros_like(x, dtype=np.float64)
    dim_len = x.shape[0]
    for i in range(dim_len):

        if N % 2 == 0:
            a, b = i - (N - 1) // 2, i + (N - 1) // 2 + 2
        else:
            a, b = i - (N - 1) // 2, i + (N - 1) // 2 + 1

        # cap indices to min and max indices
        a = max(0, a)
        b = min(dim_len, b)
        out[i] = np.mean(x[a:b])
    return out


# axial visualization and scrolling of the center points
def viewer(volume, dx, dy, center, title, textstr):
    print("center=", center)
    # remove_keymap_conflicts({'j', 'k'})
    fig = plt.figure(figsize=(12, 7))
    ax = fig.subplots()
    ax.volume = volume
    width = volume.shape[1]
    height = volume.shape[0]
    extent = (0, 0 + (volume.shape[1] * dx), 0, 0 + (volume.shape[0] * dy))
    img = ax.imshow(volume, extent=extent)
    # img=ax.imshow(volume)
    ax.set_xlabel("x distance [mm]")
    ax.set_ylabel("y distance [mm]")
    # ax.set_xlabel('x pixel')
    # ax.set_ylabel('y pixel')
    ax.set_xlim(width * dx / 2 - 10, width * dx / 2 + 10)
    ax.set_ylim(height * dy / 2 - 10, height * dy / 2 + 10)

    # fig.suptitle('Image', fontsize=16)
    print(title[0])
    ax.set_title(title[0] + "\n" + title[1], fontsize=16)
    ax.text((volume.shape[1] + 250) * dx, (volume.shape[0]) * dy, textstr)
    fig.subplots_adjust(right=0.75)
    fig.colorbar(img, ax=ax, orientation="vertical")
    # fig.canvas.mpl_connect('key_press_event', process_key_axial)

    return fig, ax


def scalingAnalysis(ArrayDicom_o, dx, dy):  # determine scaling
    ArrayDicom = u.norm01(ArrayDicom_o)
    blobs_log = blob_log(
        ArrayDicom, min_sigma=1, max_sigma=5, num_sigma=20, threshold=0.15
    )  # run on windows, for some stupid reason exclude_border is not recognized in my distro at home

    point_det = []
    for blob in blobs_log:
        y, x, r = blob
        point_det.append((x, y, r))

    point_det = sorted(
        point_det, key=itemgetter(2), reverse=True
    )  # here we sort by the radius of the dot bigger dots are around the center and edges

    point_det = np.asarray(point_det)

    # now we need to select the most extreme left and right point
    print(np.shape(ArrayDicom)[0] // 2)
    print(abs(point_det[:6, 1] - np.shape(ArrayDicom)[0] // 2) < 10)
    point_sel = []
    for i in range(0, 6):
        if abs(point_det[i, 1] - np.shape(ArrayDicom)[0] // 2) < 10:
            point_sel.append(abs(point_det[i, :]))

    point_sel = np.asarray(point_sel)

    imax = np.argmax(point_sel[:, 0])
    imin = np.argmin(point_sel[:, 0])

    print(point_sel[imax, :], point_sel[imin, :])
    distance = (
        np.sqrt(
            (point_sel[imax, 0] - point_sel[imin, 0])
            * (point_sel[imax, 0] - point_sel[imin, 0])
            * dx
            * dx
            + (point_sel[imax, 1] - point_sel[imin, 1])
            * (point_sel[imax, 1] - point_sel[imin, 1])
            * dy
            * dy
        )
        / 10.0
    )

    print("distance=", distance, "cm")  # distance is reported in cm

    # plotting the figure of scaling results

    fig = plt.figure(figsize=(12, 7))
    ax = fig.subplots()
    ax.volume = ArrayDicom_o
    width = ArrayDicom_o.shape[1]
    height = ArrayDicom_o.shape[0]
    extent = (0, 0 + (width * dx), 0, 0 + (height * dy))
    img = ax.imshow(ArrayDicom_o, extent=extent, origin="lower")
    # img = ax.imshow(ArrayDicom_o)
    ax.set_xlabel("x distance [mm]")
    ax.set_ylabel("y distance [mm]")

    ax.scatter(point_sel[imax, 0] * dx, point_sel[imax, 1] * dy)
    ax.scatter(point_sel[imin, 0] * dx, point_sel[imin, 1] * dy)
    fig.colorbar(img, ax=ax, orientation="vertical")

    # adding a horizontal arrow
    ax.annotate(
        s="",
        xy=(point_sel[imax, 0] * dx, point_sel[imax, 1] * dy),
        xytext=(point_sel[imin, 0] * dx, point_sel[imin, 1] * dy),
        arrowprops=dict(arrowstyle="<->", color="r"),
    )  # example on how to plot a double headed arrow
    ax.text(
        (width // 2.8) * dx,
        (height // 2 + 10) * dy,
        "Distance=" + str(round(distance, 4)) + " cm",
        rotation=0,
        fontsize=14,
        color="r",
    )

    return distance, fig


def full_imageProcess(ArrayDicom_o, dx, dy, title):  # process a full image
    ArrayDicom = u.norm01(ArrayDicom_o)
    height = np.shape(ArrayDicom)[0]
    width = np.shape(ArrayDicom)[1]

    blobs_log = blob_log(
        ArrayDicom, min_sigma=1, max_sigma=5, num_sigma=20, threshold=0.15
    )  # run on windows, for some stupid reason exclude_border is not recognized in my distro at home

    center = []
    point_det = []
    for blob in blobs_log:
        y, x, r = blob
        point_det.append((x, y, r))

    point_det = sorted(
        point_det, key=itemgetter(2), reverse=True
    )  # here we sort by the radius of the dot bigger dots are around the center and edges

    # we need to find the centre dot as well as the larger dots on the sides of the image

    # for j in range(0, len(point_det)):
    #     x, y, r = point_det[j]
    #     center.append((int(round(x)), int(round(y))))

    # now that we have detected the centre we are going to increase the precision of the detected point
    im_centre = Image.fromarray(
        255
        * ArrayDicom[
            height // 2 - 20 : height // 2 + 20, width // 2 - 20 : width // 2 + 20
        ]
    )
    im_centre = im_centre.resize(
        (im_centre.width * 10, im_centre.height * 10), Image.LANCZOS
    )

    xdet_int, ydet_int = point_detect_singleImage(im_centre)
    xdet = int(width // 2 - 20) + xdet_int / 10
    ydet = int(height // 2 - 20) + ydet_int / 10

    center.append((xdet, ydet))

    textstr = ""

    print("center=", center)
    fig, ax = viewer(u.range_invert(ArrayDicom_o), dx, dy, center, title, textstr)

    return fig, ax, center


def full_imageProcess_noGraph(ArrayDicom_o):  # process a full image
    ArrayDicom = u.norm01(ArrayDicom_o)
    height = np.shape(ArrayDicom)[0]
    width = np.shape(ArrayDicom)[1]

    blobs_log = blob_log(
        ArrayDicom, min_sigma=1, max_sigma=5, num_sigma=20, threshold=0.15
    )  # run on windows, for some stupid reason exclude_border is not recognized in my distro at home

    center = []
    point_det = []
    for blob in blobs_log:
        y, x, r = blob
        point_det.append((x, y, r))

    point_det = sorted(
        point_det, key=itemgetter(2), reverse=True
    )  # here we sort by the radius of the dot bigger dots are around the center and edges

    # we need to find the centre dot as well as the larger dots on the sides of the image

    # for j in range(0, len(point_det)):
    #     x, y, r = point_det[j]
    #     center.append((int(round(x)), int(round(y))))

    # now that we have detected the centre we are going to increase the precision of the detected point
    im_centre = Image.fromarray(
        255
        * ArrayDicom[
            height // 2 - 20 : height // 2 + 20, width // 2 - 20 : width // 2 + 20
        ]
    )
    im_centre = im_centre.resize(
        (im_centre.width * 10, im_centre.height * 10), Image.LANCZOS
    )

    xdet_int, ydet_int = point_detect_singleImage(im_centre)
    xdet = int(width // 2 - 20) + xdet_int / 10
    ydet = int(height // 2 - 20) + ydet_int / 10

    center.append((xdet, ydet))

    # fig, ax=viewer(u.range_invert(ArrayDicom_o), dx, dy, center, title, textstr)

    return center


def point_detect_singleImage(imcirclist):
    detCenterXRegion = []
    detCenterYRegion = []

    print("Finding bibs in phantom...")
    grey_img = np.array(imcirclist, dtype=np.uint8)  # converting the image to grayscale
    blobs_log = blob_log(
        grey_img, min_sigma=15, max_sigma=50, num_sigma=10, threshold=0.05
    )

    centerXRegion = []
    centerYRegion = []
    centerRRegion = []
    grey_ampRegion = []
    for blob in blobs_log:
        y, x, r = blob
        # center = (int(x), int(y))
        centerXRegion.append(x)
        centerYRegion.append(y)
        centerRRegion.append(r)
        grey_ampRegion.append(grey_img[int(y), int(x)])

    xindx = int(centerXRegion[np.argmin(grey_ampRegion)])
    yindx = int(centerYRegion[np.argmin(grey_ampRegion)])
    # rindx = int(centerRRegion[np.argmin(grey_ampRegion)])

    detCenterXRegion = xindx
    detCenterYRegion = yindx

    return detCenterXRegion, detCenterYRegion


# def read_dicom(filename1,filename2,ioption):
def read_dicom(directory):
    now = datetime.now()
    for subdir, dirs, files in os.walk(directory):  # pylint: disable = unused-variable
        dirs.clear()
        list_title = []
        list_gantry_angle = []
        list_collimator_angle = []
        list_figs = []
        center = []
        center_g0 = [(0, 0)]
        center_g90 = [(0, 0)]
        center_g180 = [(0, 0)]
        center_g270 = [(0, 0)]
        dx = 0
        dy = 0

        k = 0  # we callect all the images in ArrayDicom
        for file in tqdm(sorted(files)):
            print("Reading file=>", file)

            if os.path.splitext(directory + file)[1] == ".dcm":
                dataset = pydicom.dcmread(directory + file)
                gantry_angle = dataset[0x300A, 0x011E].value
                collimator_angle = dataset[0x300A, 0x0120].value

                list_gantry_angle.append(gantry_angle)
                list_collimator_angle.append(collimator_angle)

                if round(gantry_angle) == 360:
                    gantry_angle = 0
                if round(collimator_angle) == 360:
                    collimator_angle = 0

                title = (
                    "g" + str(round(gantry_angle)),
                    "c" + str(round(collimator_angle)),
                )
                print(title)

                if k == 0:
                    # title = ('Gantry= ' + str(gantry_angle), 'Collimator= ' + str(collimator_angle))
                    title = (
                        "g" + str(round(gantry_angle)),
                        "c" + str(round(collimator_angle)),
                    )
                    list_title.append(title)
                    ArrayDicom = dataset.pixel_array
                    height = np.shape(ArrayDicom)[0]
                    width = np.shape(ArrayDicom)[1]
                    SID = dataset.RTImageSID
                    dx = 1 / (SID * (1 / dataset.ImagePlanePixelSpacing[0]) / 1000)
                    dy = 1 / (SID * (1 / dataset.ImagePlanePixelSpacing[1]) / 1000)
                    print("pixel spacing row [mm]=", dx)
                    print("pixel spacing col [mm]=", dy)
                    distance, fig_scaling = scalingAnalysis(ArrayDicom, dx, dy)
                    print("distance", distance)

                else:
                    list_title.append(title)
                    tmp_array = dataset.pixel_array
                    tmp_array = u.norm01(tmp_array)
                    ArrayDicom = np.dstack((ArrayDicom, tmp_array))

                k = k + 1

    # After we colect all the images we only select g0c90 and g0c270 to calculate the center at g0

    k = 0
    l = 0
    m = 0
    n = 0

    for i, _ in enumerate(list_title):
        if list_title[i][0] == "g0" and k == 0:
            k = k + 1
            # height = np.shape(ArrayDicom[:, :, i])[0]
            # width = np.shape(ArrayDicom[:, :, i])[1]
            fig_g0c90, ax_g0c90, center_g0c90 = full_imageProcess(
                ArrayDicom[:, :, i], dx, dy, list_title[i]
            )
            center_g0[0] = (
                center_g0[0][0] + center_g0c90[0][0],
                center_g0[0][1] + center_g0c90[0][1],
            )

            list_figs.append(fig_g0c90)  # we plot always the image at g0c90

        if list_title[i][0] == "g0" and k != 0:
            k = k + 1
            center_g0c270 = full_imageProcess_noGraph(ArrayDicom[:, :, i])
            center_g0[0] = (
                center_g0[0][0] + center_g0c270[0][0],
                center_g0[0][1] + center_g0c270[0][1],
            )

    for i, _ in enumerate(list_title):
        if list_title[i][0] == "g90":
            l = l + 1
            center_g90c = full_imageProcess_noGraph(ArrayDicom[:, :, i])
            center_g90[0] = (
                center_g90[0][0] + center_g90c[0][0],
                center_g90[0][1] + center_g90c[0][1],
            )

        if list_title[i][0] == "g180":
            m = m + 1
            center_g180c = full_imageProcess_noGraph(ArrayDicom[:, :, i])
            center_g180[0] = (
                center_g180[0][0] + center_g180c[0][0],
                center_g180[0][1] + center_g180c[0][1],
            )

        if list_title[i][0] == "g270":
            n = n + 1
            center_g270c = full_imageProcess_noGraph(ArrayDicom[:, :, i])
            center_g270[0] = (
                center_g270[0][0] + center_g270c[0][0],
                center_g270[0][1] + center_g270c[0][1],
            )

    print(k, "images used for g0")
    print(l, "images used for g90")
    print(m, "images used for g180")
    print(n, "images used for g270")

    center_g0[0] = (center_g0[0][0] / k, center_g0[0][1] / k)
    center_g90[0] = (center_g90[0][0] / l, center_g90[0][1] / l)
    center_g180[0] = (center_g180[0][0] / m, center_g180[0][1] / m)
    center_g270[0] = (center_g270[0][0] / n, center_g270[0][1] / n)

    center.append(center_g0[0])
    center.append(center_g90[0])
    center.append(center_g180[0])
    center.append(center_g270[0])

    x_g0, y_g0 = center_g0[0]
    x_g90, y_g90 = center_g90[0]
    x_g180, y_g180 = center_g180[0]
    x_g270, y_g270 = center_g270[0]

    max_deltax = 0
    max_deltay = 0
    for i in range(0, len(center)):  # pylint: disable = consider-using-enumerate
        for j in range(i + 1, len(center)):
            deltax = abs(center[i][0] - center[j][0])
            deltay = abs(center[i][1] - center[j][1])
            if deltax > max_deltax:
                max_deltax = deltax
            if deltay > max_deltay:
                max_deltay = deltay

    print("Maximum delta x =", max_deltax * dx, "mm")
    print("Maximum delta y =", max_deltay * dy, "mm")

    ax_g0c90.scatter(
        x_g0 * dx, (ArrayDicom[:, :, i].shape[0] - y_g0) * dy, label="g=0"
    )  # perfect!

    ax_g0c90.scatter(
        x_g90 * dx, (ArrayDicom[:, :, i].shape[0] - y_g90) * dy, label="g=90"
    )  # perfect!
    ax_g0c90.scatter(
        x_g180 * dx, (ArrayDicom[:, :, i].shape[0] - y_g180) * dy, label="g=180"
    )  # perfect!
    ax_g0c90.scatter(
        x_g270 * dx, (ArrayDicom[:, :, i].shape[0] - y_g270) * dy, label="g=270"
    )  # perfect!

    # print(list_title[i], "center_g0c90=", center_g0c90, "center=", center, dist)
    ax_g0c90.legend(bbox_to_anchor=(1.25, 1), loc=2, borderaxespad=0.0)

    # adding a horizontal arrow
    # ax.annotate(
    #     s="",
    #     xy=(point_sel[imax, 0] * dx, point_sel[imax, 1] * dy),
    #     xytext=(point_sel[imin, 0] * dx, point_sel[imin, 1] * dy),
    #     arrowprops=dict(arrowstyle="<->", color="r"),
    # )  # example on how to plot a double headed arrow
    ax_g0c90.text(
        (width // 2.15) * dx,
        (height // 2.15) * dy,
        "Maximum delta x =" + str(round(max_deltax * dx, 4)) + " mm",
        rotation=0,
        fontsize=14,
        color="r",
    )
    ax_g0c90.text(
        (width // 2.15) * dx,
        (height // 2.18) * dy,
        "Maximum delta y =" + str(round(max_deltay * dy, 4)) + " mm",
        rotation=0,
        fontsize=14,
        color="r",
    )

    if platform == "linux":
        output_flnm = (
            dirname
            + "/"
            + now.strftime("%d-%m-%Y_%H:%M_")
            + dataset[0x0008, 0x1010].value
            + "_Graticule_report.pdf"
        )
    elif platform == "win32":
        output_flnm = dataset[0x0008, 0x1010].value + "_Graticule_report.pdf"

    # with PdfPages(directory + "/" + output_flnm) as pdf:
    with PdfPages(output_flnm) as pdf:
        # Page = plt.figure(figsize=(4, 5))
        # Page.text(0, 0.9, 'Report', size=18)
        # Page.text(0, 0.9, "Distance=" + str(distance)+ " cm", size=14)
        pdf.savefig(fig_g0c90)
        pdf.savefig(fig_scaling)

    # exit(0)
    sys.exit(0)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()  # pylint: disable = invalid-name
    parser.add_argument("-d", "--directory", help="path to folder")
    args = parser.parse_args()  # pylint: disable = invalid-name

    if args.directory:
        dirname = args.directory  # pylint: disable = invalid-name
        read_dicom(dirname)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys._imports import numpy as np


def jaw_dd2dcm(jaw):
    jaw = np.array(jaw, copy=True)
    jaw[:, 1] = -jaw[:, 1]

    converted_jaw = jaw.astype(str)
    converted_jaw[:, 1] = jaw.astype(str)[:, 0]
    converted_jaw[:, 0] = jaw.astype(str)[:, 1]
    converted_jaw = converted_jaw.tolist()

    return converted_jaw


def mlc_dd2dcm(mlc):
    mlc = np.array(mlc, copy=False)

    dicom_mlc_format = []
    for control_point in mlc:
        concatenated = np.hstack([-control_point[-1::-1, 1], control_point[-1::-1, 0]])
        dicom_mlc_format.append(concatenated.astype(str).tolist())

    return dicom_mlc_format


def angle_dd2dcm(angle):
    diff = np.append(np.diff(angle), 0)
    movement = (np.empty_like(angle)).astype(str)  # pylint: disable = no-member

    movement[diff > 0] = "CW"
    movement[diff < 0] = "CC"
    movement[diff == 0] = "NONE"

    converted_angle = np.array(angle, copy=False)
    converted_angle[converted_angle < 0] = converted_angle[converted_angle < 0] + 360

    converted_angle = converted_angle.astype(str).tolist()

    return converted_angle, movement


def gantry_tol_from_gantry_angles(gantry_angles):
    min_diff = np.min(np.diff(sorted(gantry_angles)))
    gantry_tol = np.min([min_diff / 2 - 0.1, 3])

    return gantry_tol



# Copyright (C) 2020 University of New South Wales & Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=redefined-outer-name

import shutil
import tempfile
import pathlib
import time
from unittest.mock import Mock

import pytest

import pydicom
from pynetdicom import AE, VerificationPresentationContexts
from pynetdicom.sop_class import RTPlanStorage  # pylint: disable=no-name-in-module

from pymedphys._dicom.connect.listen import DicomListener
from pymedphys._dicom.create import dicom_dataset_from_dict

from pymedphys._utilities.test import process

# TODO How to determine an appropriate port for testing?
TEST_PORT = 9988

METHOD_MOCK = Mock()


@pytest.fixture()
def listener():

    dicom_listener = DicomListener(
        port=TEST_PORT, on_released_callback=METHOD_MOCK.method
    )
    dicom_listener.start()

    yield dicom_listener

    dicom_listener.stop()


@pytest.mark.pydicom
def test_dicom_listener_echo(listener):
    """Test to ensure that running dicom listener responds to C-ECHO
    """

    # Send C-ECHO
    ae = AE()
    ae.requested_contexts = VerificationPresentationContexts

    assoc = ae.associate(listener.host, listener.port, ae_title=listener.ae_title)

    result = None
    if assoc.is_established:
        status = assoc.send_c_echo()

        if status:
            result = status.Status

        assoc.release()

    # Check we got a valid result
    assert result == 0


@pytest.fixture()
def test_dataset():
    """pytest fixture to returna a dummy DICOM dataset for testing

    Returns
    -------
    test_dataset : ``pydicom.dataset.Dataset``
        A dummy DICOM dataset
    """

    # Create a test Dicom object
    test_uid = pydicom.uid.generate_uid()
    test_series_uid = pydicom.uid.generate_uid()
    test_dataset = dicom_dataset_from_dict(
        {
            "SOPClassUID": RTPlanStorage,
            "SOPInstanceUID": test_uid,
            "SeriesInstanceUID": test_series_uid,
            "Modality": "RTPlan",
            "Manufacturer": "PyMedPhys",
            "BeamSequence": [{"Manufacturer": "PyMedPhys"}],
        }
    )

    file_meta = pydicom.dataset.FileMetaDataset(
        dicom_dataset_from_dict(
            {
                "FileMetaInformationVersion": "01",
                "MediaStorageSOPClassUID": RTPlanStorage,
                "MediaStorageSOPInstanceUID": test_uid,
                "TransferSyntaxUID": pydicom.uid.ImplicitVRLittleEndian,
                "ImplementationClassUID": pydicom.uid.PYDICOM_IMPLEMENTATION_UID,
                "ImplementationVersionName": "PYMEDPHYSDCM",
            }
        )
    )

    test_dataset.file_meta = file_meta
    test_dataset.is_implicit_VR = True
    test_dataset.is_little_endian = True
    test_dataset.fix_meta_info(enforce_standard=True)

    return test_dataset


@pytest.mark.pydicom
def test_dicom_listener_send(listener, test_dataset):
    """Test to ensure that running dicom listener receives a stores a Dicom file
    """

    METHOD_MOCK.reset_mock()

    # Send the data to the listener
    ae = AE()
    ae.add_requested_context(RTPlanStorage)
    assoc = ae.associate(listener.host, listener.port, ae_title="PYMEDPHYSTEST")
    assert assoc.is_established
    status = assoc.send_c_store(test_dataset)
    assert status.Status == 0
    assoc.release()

    # Check that it was received
    METHOD_MOCK.method.assert_called_once()
    args, _ = METHOD_MOCK.method.call_args_list[0]
    storage_path = args[0]
    file_path = storage_path / f"RP.{test_dataset.SOPInstanceUID}"
    assert file_path.exists()

    read_dataset = pydicom.read_file(file_path)
    assert read_dataset.SeriesInstanceUID == test_dataset.SeriesInstanceUID

    # Clean up after ourselves
    shutil.rmtree(storage_path)


@pytest.mark.pydicom
def test_dicom_listener_send_conflicting_file(listener, test_dataset):
    """Test to ensure that running dicom listener receives a stores a Dicom file
    """

    METHOD_MOCK.reset_mock()

    # Send the data to the listener
    ae = AE()
    ae.add_requested_context(RTPlanStorage)
    assoc = ae.associate(listener.host, listener.port, ae_title="PYMEDPHYSTEST")
    assert assoc.is_established
    status = assoc.send_c_store(test_dataset)
    assert status.Status == 0
    assoc.release()

    # Send again, should succeed without writing the same file again
    ae = AE()
    ae.add_requested_context(RTPlanStorage)
    assoc = ae.associate(listener.host, listener.port, ae_title="PYMEDPHYSTEST")
    assert assoc.is_established
    status = assoc.send_c_store(test_dataset)
    assert status.Status == 0
    assoc.release()

    # Modify the file to make it conflict
    args, _ = METHOD_MOCK.method.call_args_list[0]
    storage_path = args[0]
    file_path = storage_path / f"RP.{test_dataset.SOPInstanceUID}"
    ds = pydicom.read_file(file_path)
    ds.Manufacturer = "PyMedPhysModified"
    ds.save_as(file_path, write_like_original=False)

    # Send again, should save the file in the orphan directory
    ae = AE()
    ae.add_requested_context(RTPlanStorage)
    assoc = ae.associate(listener.host, listener.port, ae_title="PYMEDPHYSTEST")
    assert assoc.is_established
    status = assoc.send_c_store(test_dataset)
    assert status.Status == 0
    assoc.release()

    # Check the original file is the same
    read_dataset = pydicom.read_file(file_path)
    assert read_dataset.Manufacturer == "PyMedPhysModified"

    # Check the other file was written to the orphan directory
    orphan_files = list(storage_path.joinpath("orphan").glob("*"))
    assert len(orphan_files) == 1
    orphan_dataset = pydicom.read_file(orphan_files[0])
    assert orphan_dataset.Manufacturer == "PyMedPhys"

    # Clean up after ourselves
    shutil.rmtree(storage_path)


@pytest.mark.pydicom
def test_dicom_listener_cli(test_dataset):

    test_directory = pathlib.Path(tempfile.mkdtemp())

    scp_ae_title = "PYMEDPHYSTEST"

    with process(
        f"pymedphys dicom listen {TEST_PORT} -d {test_directory} -a {scp_ae_title}",
        shell=True,
    ) as _:

        # Send the data to the listener
        ae = AE()
        ae.add_requested_context(RTPlanStorage)
        assoc = ae.associate("127.0.0.1", TEST_PORT, ae_title=scp_ae_title)

        # Give the process a few seconds to start up
        elapsed = 0
        while not assoc.is_established:
            time.sleep(0.5)
            elapsed += 0.5
            if elapsed >= 3:  # Break if still not connecting after 3 seconds
                break

            assoc = ae.associate("127.0.0.1", TEST_PORT, ae_title=scp_ae_title)
        assert assoc.is_established
        status = assoc.send_c_store(test_dataset)
        assert status.Status == 0
        assoc.release()

    series_dir = test_directory.joinpath(test_dataset.SeriesInstanceUID)
    file_path = series_dir.joinpath(f"RP.{test_dataset.SOPInstanceUID}")
    read_dataset = pydicom.read_file(file_path)
    assert read_dataset.SeriesInstanceUID == test_dataset.SeriesInstanceUID

    # Clean up after ourselves
    shutil.rmtree(test_directory)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pathlib

import pytest

import numpy as np
import pandas as pd

import pymedphys
import pymedphys._wlutz.iview

HERE = pathlib.Path(__file__).parent.resolve()


@pytest.mark.slow
def test_end_to_end():
    edge_lengths = [20, 20]

    image_paths = pymedphys.zip_data_paths("wlutz_images.zip")
    results = pymedphys._wlutz.iview.batch_process(  # pylint:disable = protected-access
        image_paths, edge_lengths, display_figure=False
    )

    reference_dataframe = pd.read_csv(HERE.joinpath("end_to_end.csv"))

    assert np.max(np.abs(results["Rotation"] - reference_dataframe["Rotation"])) <= 0.1
    assert (
        np.max(
            np.max(
                np.abs(
                    results.drop(columns=["Rotation"])
                    - reference_dataframe.drop(columns=["Rotation"])
                )
            )
        )
        <= 0.05
    )



# Copyright (C) 2016 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys._imports import numpy as np
from pymedphys._imports import scipy

from . import api, normalisation


def mephysto_absolute_depth_dose(
    absolute_dose, depth_of_absolute_dose_mm, distance, relative_dose, scan_curvetype
):
    choose_mephysto = scan_curvetype == "PDD"
    if np.sum(choose_mephysto) != 1:
        raise ValueError("Can only handle one PDD per mephysto file")

    mephysto_pdd_depth = distance[choose_mephysto][0]
    mephysto_dose = relative_dose[choose_mephysto][0]

    interpolation = scipy.interpolate.interp1d(mephysto_pdd_depth, mephysto_dose)

    mephysto_pdd_dose = (
        mephysto_dose / interpolation(depth_of_absolute_dose_mm) * absolute_dose
    )

    return mephysto_pdd_depth, mephysto_pdd_dose


def mephysto_absolute_profiles(
    curvetype,
    depth_test,
    distance,
    relative_dose,
    scan_curvetype,
    scan_depth,
    mephysto_pdd_depth,
    mephysto_pdd_dose,
):

    choose_mephysto = (scan_curvetype == curvetype) & (scan_depth == depth_test)

    if np.sum(choose_mephysto) != 1:
        raise ValueError("Can only handle exactly one scan type per mephysto file")

    mephysto_distance = distance[choose_mephysto][0]
    mephysto_normalised_dose = normalisation.normalise_profile(
        mephysto_distance,
        relative_dose[choose_mephysto][0],
        scale_to_pdd=True,
        pdd_distance=mephysto_pdd_depth,
        pdd_relative_dose=mephysto_pdd_dose,
        scan_depth=depth_test,
    )

    return mephysto_distance, mephysto_normalised_dose


def absolute_scans_from_mephysto(
    mephysto_file, absolute_dose, depth_of_absolute_dose_mm
):
    distance, relative_dose, scan_curvetype, scan_depth = api.load_mephysto(
        mephysto_file
    )

    depth_testing = scan_depth[~np.isnan(scan_depth)]
    depth_testing = np.unique(depth_testing)

    if depth_of_absolute_dose_mm == "dmax":
        choose_mephysto = scan_curvetype == "PDD"
        mephysto_pdd_depth = distance[choose_mephysto][0]
        mephysto_dose = relative_dose[choose_mephysto][0]
        depth_of_absolute_dose_mm = mephysto_pdd_depth[np.argmax(mephysto_dose)]

    mephysto_pdd_depth, mephysto_pdd_dose = mephysto_absolute_depth_dose(
        absolute_dose,
        depth_of_absolute_dose_mm,
        distance,
        relative_dose,
        scan_curvetype,
    )

    scans = {
        "depth_dose": {"displacement": mephysto_pdd_depth, "dose": mephysto_pdd_dose},
        "profiles": {},
    }

    for depth_test in depth_testing:
        (
            mephysto_distance_inplane,
            mephysto_normalised_dose_inplane,
        ) = mephysto_absolute_profiles(
            "INPLANE_PROFILE",
            depth_test,
            distance,
            relative_dose,
            scan_curvetype,
            scan_depth,
            mephysto_pdd_depth,
            mephysto_pdd_dose,
        )

        (
            mephysto_distance_crossplane,
            mephysto_normalised_dose_crossplane,
        ) = mephysto_absolute_profiles(
            "CROSSPLANE_PROFILE",
            depth_test,
            distance,
            relative_dose,
            scan_curvetype,
            scan_depth,
            mephysto_pdd_depth,
            mephysto_pdd_dose,
        )

        scans["profiles"][depth_test] = {
            "inplane": {
                "displacement": mephysto_distance_inplane,
                "dose": mephysto_normalised_dose_inplane,
            },
            "crossplane": {
                "displacement": mephysto_distance_crossplane,
                "dose": mephysto_normalised_dose_crossplane,
            },
        }

    return scans



from pymedphys._dev import docs, propagate, tests


def dev_cli(subparsers):
    dev_parser = subparsers.add_parser("dev")
    dev_subparsers = dev_parser.add_subparsers(dest="dev")
    add_docs_parser(dev_subparsers)
    add_test_parser(dev_subparsers)
    add_propagate_parser(dev_subparsers)

    return dev_parser


def add_docs_parser(dev_subparsers):
    parser = dev_subparsers.add_parser("docs")

    parser.add_argument("--live", help="Make the docs live reload", action="store_true")
    parser.add_argument("--output", help="Custom output directory for the built docs.")

    parser.set_defaults(func=docs.build_docs)


def add_test_parser(test_subparsers):
    parser = test_subparsers.add_parser("tests")
    parser.set_defaults(func=tests.run_tests)


def add_propagate_parser(test_subparsers):
    parser = test_subparsers.add_parser("propagate")
    parser.set_defaults(func=propagate.propagate_all)



# pylint: disable = unused-import

from pymedphys._experimental.fileformats.mapcheck import (
    read_mapcheck_txt as read_mapcheck,
)
from pymedphys._experimental.fileformats.mephysto import load_mephysto_directory
from pymedphys._experimental.fileformats.profiler import read_prs



# Copyright (C) 2020 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from pymedphys._imports import numpy as np

import pymedphys
import pymedphys._mocks.wlutz as wlutz_mocks
import pymedphys._wlutz.reporting as reporting


def test_normal_bb():
    field_centre = [0, 0]
    field_side_lengths = [20, 24]
    field_penumbra = 2
    field_rotation = 20

    bb_centre = [2, 2]
    bb_diameter = 8
    bb_max_attenuation = 0.3

    run_test(
        field_centre,
        field_side_lengths,
        field_penumbra,
        field_rotation,
        bb_centre,
        bb_diameter,
        bb_max_attenuation,
    )


@pytest.mark.skip(reason="This test currently demonstrates a limitation of the library")
def test_small_bb():
    field_centre = [0, 0]
    field_side_lengths = [20, 24]
    field_penumbra = 2
    field_rotation = 20

    bb_centre = [2, 2]
    bb_diameter = 3
    bb_max_attenuation = 0.3

    run_test(
        field_centre,
        field_side_lengths,
        field_penumbra,
        field_rotation,
        bb_centre,
        bb_diameter,
        bb_max_attenuation,
    )


def create_test_image(
    field_centre,
    field_side_lengths,
    field_penumbra,
    field_rotation,
    bb_centre,
    bb_diameter,
    bb_max_attenuation,
):
    field = wlutz_mocks.create_field_with_bb_func(
        field_centre,
        field_side_lengths,
        field_penumbra,
        field_rotation,
        bb_centre,
        bb_diameter,
        bb_max_attenuation,
    )

    x = np.arange(-20, 20.1, 0.1)
    y = np.arange(-22, 22.1, 0.1)
    xx, yy = np.meshgrid(x, y)

    img = field(xx, yy)

    return x, y, img


def run_test(
    field_centre,
    field_side_lengths,
    field_penumbra,
    field_rotation,
    bb_centre,
    bb_diameter,
    bb_max_attenuation,
):

    x, y, img = create_test_image(
        field_centre,
        field_side_lengths,
        field_penumbra,
        field_rotation,
        bb_centre,
        bb_diameter,
        bb_max_attenuation,
    )

    (
        determined_bb_centre,
        determined_field_centre,
        determined_field_rotation,
    ) = pymedphys.wlutz.find_field_and_bb(
        x,
        y,
        img,
        field_side_lengths,
        bb_diameter,
        penumbra=field_penumbra,
        pylinac_tol=None,
    )

    try:
        assert np.allclose(bb_centre, determined_bb_centre, atol=0.001)
        assert np.allclose(field_centre, determined_field_centre, atol=0.001)
        assert np.allclose(field_rotation, determined_field_rotation, atol=0.01)

    except:
        reporting.image_analysis_figure(
            x,
            y,
            img,
            determined_bb_centre,
            determined_field_centre,
            determined_field_rotation,
            bb_diameter,
            field_side_lengths,
            field_penumbra,
        )
        raise



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

# pylint: disable = C0413, C0103

import pymedphys


def setup(app):
    app.add_stylesheet("style.css")


# -- Project information -----------------------------------------------------


project = "PyMedPhys"
copyright = (  # pylint: disable = redefined-builtin
    "2015-2020, the PyMedPhys contributors"
)

author = "PyMedPhys contributors"

# The short X.Y version
# Currently being used to create mybinder link -- potentially need to rethink
# this logic.
version = "main"
# The full version, including alpha/beta/rc tags
release = pymedphys.__version__

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.graphviz",
    "sphinxarg.ext",
    "matplotlib.sphinxext.plot_directive",
    "sphinxcontrib.napoleon",
    "nbsphinx",
]

napoleon_google_docstring = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst"]
# source_suffix = '.rst'

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "**.ipynb_checkpoints",
    "**.dcm",
    "drafts",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# This is processed by Jinja2 and inserted before each notebook
nbsphinx_prolog = r"""
{% set docname = env.doc2path(env.docname, base='docs').replace('\\', '/') %}

.. only:: html

    .. role:: raw-html(raw)

        :format: html

    .. nbinfo::

        This page was generated from `/lib/pymedphys/{{ docname }}`__.
        Interactive online version:
        :raw-html:`<a href="https://colab.research.google.com/github/pymedphys/pymedphys/blob/{{ env.config.version }}/lib/pymedphys/{{ docname }}"><img alt="Open in colab" src="https://colab.research.google.com/assets/colab-badge.svg" style="vertical-align:text-bottom"></a>`

    __ https://github.com/pymedphys/pymedphys/blob/
        {{ env.config.version }}/lib/pymedphys/{{ docname }}
"""

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"analytics_id": "UA-26591325-7"}

html_logo = "img/pymedphys.svg"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

html_context = {
    "display_github": True,
    "github_user": "pymedphys",
    "github_repo": "pymedphys",
    "github_version": "main/lib/pymedphys/docs/",
}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "pymedphysdoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {  # type: ignore
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "pymedphys.tex", "pymedphys Documentation", author, "manual")
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "pymedphys", "pymedphys Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "pymedphys",
        "pymedphys Documentation",
        author,
        "pymedphys",
        "One line description of project.",
        "Miscellaneous",
    )
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["doc/search.html"]


# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import pathlib

from pymedphys._imports import plt
from pymedphys._imports import streamlit as st

from pymedphys import _losslessjpeg as lljpeg
from pymedphys._wlutz import findbb, findfield, imginterp, iview, reporting

# st.title("Winston Lutz Arc")

DATAEXCHANGE_DIRECTORY = pathlib.Path(r"S:\DataExchange")
IVIEW_DB = DATAEXCHANGE_DIRECTORY.joinpath("iViewDB")
IMAGES_DIR = IVIEW_DB.joinpath("img")


@st.cache()
def get_file_list():
    return tuple([str(path) for path in IMAGES_DIR.glob("*.jpg")])


@st.cache()
def read_image(path):
    return lljpeg.imread(path)


def main():
    st.write("## Parameters")

    number_of_images = st.number_input("Number of Images", 10)

    width = st.number_input("Width (mm)", 20)
    length = st.number_input("Length (mm)", 24)
    edge_lengths = [width, length]

    # initial_rotation = 0
    bb_diameter = st.number_input("BB Diameter (mm)", 8)
    penumbra = st.number_input("Penumbra (mm)", 2)

    # files = sorted(IMAGES_DIR.glob("*.jpg"), key=lambda t: -os.stat(t).st_mtime)
    # most_recent = files[0:5]

    # most_recent

    files = get_file_list()
    most_recent = sorted(files, key=lambda t: -os.stat(t).st_mtime)[0:number_of_images]

    image_path = st.radio("Image to select", options=most_recent)

    plt.imshow(read_image(image_path))
    st.pyplot()

    if st.button("Calculate"):
        img = read_image(image_path)
        x, y, img = iview.iview_image_transform(img)
        field = imginterp.create_interpolated_field(x, y, img)
        initial_centre = findfield.get_centre_of_mass(x, y, img)
        (field_centre, field_rotation) = findfield.field_centre_and_rotation_refining(
            field, edge_lengths, penumbra, initial_centre, fixed_rotation=0
        )

        bb_centre = findbb.optimise_bb_centre(
            field, bb_diameter, edge_lengths, penumbra, field_centre, field_rotation
        )
        fig = reporting.image_analysis_figure(
            x,
            y,
            img,
            bb_centre,
            field_centre,
            field_rotation,
            bb_diameter,
            edge_lengths,
            penumbra,
        )

        st.write(fig)
        st.pyplot()



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""This module exposes a CLI that is designed to be run once per day. It pulls
diagnostic zip files from the backups on the network share on the Linac.
This diagnostic zip file contains the Elekta trf logfiles within it.

This zip file is then extracted into the to be indexed directory, and indexing
is run.
"""

import json
import os

from pymedphys._imports import pandas as pd

from .diagnostics_zips import (
    extract_diagnostic_zips_and_archive,
    fetch_system_diagnostics_multi_linac,
)
from .index import index_logfiles


def orchestration(mosaiq_sql, linac_details, logfile_data_directory):
    """Accepts a data directory for organising the log files as well as a
    machine map.

    Example
    -------

    mosaiq_sql = {
        "rccc": {
            "timezone": "Australia/Sydney",
            "mosaiq_sql_server": "mosaiq:1433"
        }
    }

    linac_details = {
        "2619": {
            "centre": "rccc",
            "ip": "10.0.0.1"
        },
        "2694": {
            "centre": "rccc",
            "ip": "10.0.0.2"
        }
    }

    data_directory = "path/to/logfile/storage"

    orchestration(mosaiq_sql, linac_details, data_directory)
    """

    machine_ip_map = {
        machine: machine_lookup["ip"]
        for machine, machine_lookup in linac_details.items()
    }

    diagnostics_directory = os.path.join(logfile_data_directory, "diagnostics")

    print("Fetching diagnostic zip files from Linacs...")
    fetch_system_diagnostics_multi_linac(machine_ip_map, diagnostics_directory)

    print("Extracting trf logfiles from diagnostics zip files...")
    extract_diagnostic_zips_and_archive(logfile_data_directory)

    print("Indexing logfiles...")
    index_logfiles(mosaiq_sql, linac_details, logfile_data_directory)


def orchestration_cli(args):
    data_directory = args.data_directory

    if args.mosaiq_sql is None:
        mosaiq_sql_path = os.path.join(data_directory, "config_mosaiq_sql.csv")
    else:
        mosaiq_sql_path = args.mosaiq_sql

    if args.linac_details is None:
        linac_details_path = os.path.join(data_directory, "config_linac_details.csv")

    else:
        linac_details_path = args.linac_details

    mosaiq_sql_table = pd.read_csv(mosaiq_sql_path, index_col=0)
    linac_details_table = pd.read_csv(linac_details_path, index_col=0)

    print("Data directory used:\n    {}\n".format(data_directory))

    mosaiq_sql = {
        str(centre): {
            "timezone": row["Timezone"],
            "mosaiq_sql_server": row["Mosaiq SQL Server (Hostname:Port)"],
        }
        for centre, row in mosaiq_sql_table.iterrows()
    }

    print(
        "Mosaiq SQL configuration used:\n{}\n".format(json.dumps(mosaiq_sql, indent=4))
    )

    linac_details = {
        str(machine): {"centre": row["Centre"], "ip": row["IP"]}
        for machine, row in linac_details_table.iterrows()
    }

    print("Linac configuration used:\n{}\n".format(json.dumps(linac_details, indent=4)))

    orchestration(mosaiq_sql, linac_details, data_directory)



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Notes

Make the prescan be the moving image, that way the treatment field is
not interpolated during the netOD calc and is left in the same coords as
the laser marked image.

"""
# The following needs to be removed before leaving the experimental module
# pylint: skip-file

import pytest
from fixtures import postscans, prescans

from pymedphys._experimental.film import calc_net_od

# import numpy as np
# import matplotlib.pyplot as plt


def test_net_od(prescans, postscans):  # pylint: disable=redefined-outer-name
    keys = prescans.keys()
    assert keys == postscans.keys()

    results = {}

    keys_to_use = keys
    keys_to_use = [200, 600]

    for key in keys_to_use:
        results[key] = calc_net_od(prescans[key], postscans[key])
        # plt.pcolormesh(results[key])
        # plt.colorbar()
        # plt.axis('equal')

        # plt.show()



# Copyright (C) 2019-2020 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys import Delivery

# pylint: disable = protected-access


def test_object_consistency():
    empty = Delivery._empty()
    filtered = empty._filter_cps()

    assert isinstance(filtered.monitor_units, tuple)

    filtered._metersets(0, 0)


def test_base_object():
    empty = Delivery._empty()

    assert empty.monitor_units == tuple()

    collection = {field: getattr(empty, field) for field in empty._fields}

    dummy = Delivery(**collection)



# Copyright (C) 2019 Cancer Care Associates and Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import pydicom

import pymedphys

DIRS_TO_SKIP = (
    "10x10",  #  3DCRT fields not yet implemented within Delivery.from_monaco
)


@pytest.mark.pydicom
def test_delivery_from_monaco():
    data_paths = pymedphys.zip_data_paths("tel-dicom-pairs.zip")
    dir_names = {path.parent.name for path in data_paths}

    assert len(dir_names) >= 2

    for dir_name in dir_names:
        if dir_name in DIRS_TO_SKIP:
            continue

        current_paths = [path for path in data_paths if path.parent.name == dir_name]

        tel_path = get_file_type(current_paths, "tel.1", exact_match=True)
        dcm_path = get_file_type(current_paths, "dcm")

        delivery_dcm = pymedphys.Delivery.from_dicom(
            pydicom.read_file(str(dcm_path), force=True), fraction_number=1
        )

        delivery_monaco = pymedphys.Delivery.from_monaco(tel_path)

        assert np.allclose(delivery_monaco.mu, delivery_dcm.mu, atol=0.01)
        assert np.allclose(delivery_monaco.gantry, delivery_dcm.gantry, atol=0.01)
        assert np.allclose(
            delivery_monaco.collimator, delivery_dcm.collimator, atol=0.01
        )
        assert np.allclose(delivery_monaco.mlc, delivery_dcm.mlc, atol=0.1)
        assert np.allclose(delivery_monaco.jaw, delivery_dcm.jaw, atol=0.01)


def get_file_type(input_paths, name, exact_match=False):
    if exact_match:
        paths = [path for path in input_paths if name == path.name]
    else:
        paths = [path for path in input_paths if name in path.name]
    assert len(paths) == 1
    return paths[0]



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from pymedphys._experimental.film import create_dose_function


def test_create_dose_function():
    net_od = [
        0.03321279,
        0.16718541,
        0.21453297,
        0.25253895,
        0.28419923,
        0.30208024,
        0.3075344,
        0.31008735,
        0.3157614,
        0.31611426,
        0.32567696,
        0.33727732,
        0.34633589,
        0.36541121,
        0.37880263,
        0.39419754,
        0.4152226,
        0.43042048,
    ]

    dose = [
        0.0,
        200.0,
        300.0,
        400.0,
        500.0,
        550.0,
        580.0,
        600.0,
        650.0,
        620.0,
        700.0,
        750.0,
        800.0,
        900.0,
        1000.0,
        1100.0,
        1250.0,
        1400.0,
    ]

    baseline_fitted_dose = [
        31.12929272,
        193.62365785,
        289.34866228,
        392.43366586,
        501.12974266,
        573.27868306,
        596.98144741,
        608.35995856,
        634.31168013,
        635.95616367,
        681.91912861,
        741.4117247,
        790.84547747,
        903.97412589,
        991.13790282,
        1099.74637299,
        1263.53372935,
        1393.75250018,
    ]

    dose_function = create_dose_function(net_od, dose)
    assert np.allclose(baseline_fitted_dose, dose_function(net_od), 0.0001, 0.0001)



# Copyright (C) 2018 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Index logfiles.
"""

import json
import os
import pathlib
import traceback
from glob import glob

from pymedphys._imports import attr

from pymedphys._mosaiq.connect import multi_mosaiq_connect
from pymedphys._mosaiq.delivery import NoMosaiqEntries, get_mosaiq_delivery_details
from pymedphys._trf.decode.header import Header, decode_header_from_file
from pymedphys._utilities.filehash import hash_file
from pymedphys._utilities.filesystem import make_a_valid_directory_name

from .identify import date_convert


def create_logfile_directory_name(
    centre, delivery_details, header: Header, path_string_time
):
    proposed_patient_directory_name = "{}_{}_{}".format(
        delivery_details.patient_id,
        delivery_details.last_name,
        delivery_details.first_name,
    )

    patient_directory_name = make_a_valid_directory_name(
        proposed_patient_directory_name
    )

    if delivery_details.qa_mode:
        clinical_or_qa = "qa"
    else:
        clinical_or_qa = "clinical"

    proposed_field_directory_name = "{}_{}_{}_{}".format(
        delivery_details.field_id,
        header.field_label,
        header.field_name,
        delivery_details.field_type,
    )

    field_directory_name = make_a_valid_directory_name(proposed_field_directory_name)

    proposed_delivery_directory = "{}_{}".format(path_string_time, header.machine)

    delivery_directory = make_a_valid_directory_name(proposed_delivery_directory)

    return os.path.join(
        centre,
        patient_directory_name,
        clinical_or_qa,
        field_directory_name,
        delivery_directory,
    )


def create_index_entry(
    new_filepath, delivery_details, header: Header, mosaiq_string_time
):

    index_entry = {
        "filepath": new_filepath,
        "delivery_details": {**attr.asdict(delivery_details)},
        "logfile_header": {**header._asdict()},
        "local_time": mosaiq_string_time,
    }

    return index_entry


# CODE SMELL: The logic within `file_already_in_index` is actually
# already handled by `rename_and_handle_fileexists`. This is
# unnecessarily duplicated.


def file_already_in_index(indexed_filepath, to_be_indexed_filepath, filehash):
    try:
        new_hash = hash_file(indexed_filepath)
    except FileNotFoundError:
        raise FileNotFoundError(
            "Indexed logfile can't be found in its declared location."
        )

    try:
        assert new_hash == filehash
    except AssertionError:
        raise AssertionError("The located file doesn't agree with the index hash.")

    print("Already exists in index")
    os.remove(to_be_indexed_filepath)


def rename_and_handle_fileexists(old_filepath, new_filepath):
    pathlib.Path(new_filepath).parent.mkdir(parents=True, exist_ok=True)

    try:
        os.rename(old_filepath, new_filepath)
    except FileExistsError:
        files_hashes_match = hash_file(old_filepath) == hash_file(new_filepath)
        print("    File already exists at {}".format(new_filepath))
        if files_hashes_match:
            os.remove(old_filepath)
        else:
            raise FileExistsError("File already exists and the hash does not match")


def get_logfile_mosaiq_info(
    cursor, machine_id, utc_date, mosaiq_timezone, field_label, field_name, buffer=240
):
    mosaiq_string_time, _ = date_convert(utc_date, mosaiq_timezone)
    delivery_details = get_mosaiq_delivery_details(
        cursor, machine_id, mosaiq_string_time, field_label, field_name, buffer=buffer
    )

    return attr.asdict(delivery_details)


# TODO Split this function up into smaller functions for easier reuse.
def file_ready_to_be_indexed(
    cursors,
    filehash_list,
    to_be_indexed_dict,
    unknown_error_in_logfile,
    no_mosaiq_record_found,
    no_field_label_in_logfile,
    indexed_directory,
    index_filepath,
    index,
    machine_map,
    centre_details,
    centre_server_map,
):
    for filehash in filehash_list:
        logfile_basename = os.path.basename(to_be_indexed_dict[filehash])

        try:
            header = decode_header_from_file(to_be_indexed_dict[filehash])
            print("\n{}".format(header))
            if header.field_label == "":
                print("No field label in logfile")
                new_filepath = os.path.join(no_field_label_in_logfile, logfile_basename)
                rename_and_handle_fileexists(to_be_indexed_dict[filehash], new_filepath)
                continue

            centre = machine_map[header.machine]["centre"]
            server = centre_server_map[centre]

            mosaiq_string_time, path_string_time = date_convert(
                header.date, centre_details[centre]["timezone"]
            )
        except Exception as e:  # pylint: disable = broad-except
            traceback.print_exc()
            new_filepath = os.path.join(unknown_error_in_logfile, logfile_basename)
            rename_and_handle_fileexists(to_be_indexed_dict[filehash], new_filepath)
            continue

        try:
            delivery_details = get_mosaiq_delivery_details(
                cursors[server],
                header.machine,
                mosaiq_string_time,
                header.field_label,
                header.field_name,
                buffer=240,
            )
        except NoMosaiqEntries as e:
            print(e)
            new_filepath = os.path.join(no_mosaiq_record_found, logfile_basename)
            rename_and_handle_fileexists(to_be_indexed_dict[filehash], new_filepath)
            continue

        logfile_directory_name = create_logfile_directory_name(
            centre, delivery_details, header, path_string_time
        )

        abs_logfile_directory_name = os.path.abspath(
            os.path.join(indexed_directory, logfile_directory_name)
        )

        pathlib.Path(abs_logfile_directory_name).mkdir(parents=True, exist_ok=True)

        new_filepath = os.path.join(logfile_directory_name, logfile_basename)

        index[filehash] = create_index_entry(
            new_filepath, delivery_details, header, mosaiq_string_time
        )

        abs_new_filepath = os.path.abspath(
            os.path.join(indexed_directory, new_filepath)
        )

        index_extension_split = os.path.splitext(index_filepath)
        temp_index_path = "{}_temp{}".format(
            index_extension_split[0], index_extension_split[1]
        )

        with open(temp_index_path, "w") as json_data_file:
            json.dump(index, json_data_file, indent=2)

        os.replace(temp_index_path, index_filepath)

        os.rename(to_be_indexed_dict[filehash], abs_new_filepath)

        print(
            "Indexed logfile:\n    {} -->\n    {}".format(
                to_be_indexed_dict[filehash], abs_new_filepath
            )
        )


def index_logfiles(centre_map, machine_map, logfile_data_directory):
    data_directory = logfile_data_directory
    index_filepath = os.path.abspath(os.path.join(data_directory, "index.json"))
    to_be_indexed_directory = os.path.abspath(
        os.path.join(data_directory, "to_be_indexed")
    )
    indexed_directory = os.path.abspath(os.path.join(data_directory, "indexed"))
    no_mosaiq_record_found = os.path.abspath(
        os.path.join(data_directory, "no_mosaiq_record_found")
    )
    unknown_error_in_logfile = os.path.abspath(
        os.path.join(data_directory, "unknown_error_in_logfile")
    )
    no_field_label_in_logfile = os.path.abspath(
        os.path.join(data_directory, "no_field_label_in_logfile")
    )
    # machine_map = config['machine_map']
    centre_details = centre_map

    centre_server_map = {
        centre: centre_lookup["mosaiq_sql_server"]
        for centre, centre_lookup in centre_map.items()
    }

    sql_server_and_ports = [
        "{}".format(details["mosaiq_sql_server"])
        for _, details in centre_details.items()
    ]

    try:
        with open(index_filepath, "r") as json_data_file:
            index = json.load(json_data_file)
    except FileNotFoundError:
        index = {}

    indexset = set(index.keys())

    print("\nConnecting to Mosaiq SQL servers...")
    with multi_mosaiq_connect(sql_server_and_ports) as cursors:

        print("Globbing index directory...")
        to_be_indexed = glob(
            os.path.join(to_be_indexed_directory, "**/*.trf"), recursive=True
        )

        chunk_size = 50
        number_to_be_indexed = len(to_be_indexed)
        to_be_indexed_chunked = [
            to_be_indexed[i : i + chunk_size]
            for i in range(0, number_to_be_indexed, chunk_size)
        ]

        for i, a_to_be_indexed_chunk in enumerate(to_be_indexed_chunked):
            print(
                "\nHashing a chunk of logfiles ({}/{})".format(
                    i + 1, len(to_be_indexed_chunked)
                )
            )
            hashlist = [
                hash_file(filename, dot_feedback=True)
                for filename in a_to_be_indexed_chunk
            ]

            print(" ")

            to_be_indexed_dict = dict(zip(hashlist, a_to_be_indexed_chunk))

            hashset = set(hashlist)

            for filehash in list(hashset.intersection(indexset)):
                file_already_in_index(
                    os.path.join(indexed_directory, index[filehash]["filepath"]),
                    to_be_indexed_dict[filehash],
                    filehash,
                )

            file_ready_to_be_indexed(
                cursors,
                list(hashset.difference(indexset)),
                to_be_indexed_dict,
                unknown_error_in_logfile,
                no_mosaiq_record_found,
                no_field_label_in_logfile,
                indexed_directory,
                index_filepath,
                index,
                machine_map,
                centre_details,
                centre_server_map,
            )
    print("Complete")



# pylint: disable = unused-import

from ._wlutz.core import find_field, find_field_and_bb
from ._wlutz.iview import iview_image_transform
from ._wlutz.reporting import image_analysis_figure as reporting



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pathlib

from pymedphys._imports import toml

is_cli = False


def get_config_dir():
    config_dir = pathlib.Path.home().joinpath(".pymedphys")
    config_dir.mkdir(exist_ok=True)

    return config_dir


def get_config(path=None):
    if path is None:
        path = get_config_dir()

    path = pathlib.Path(path)

    config_path = path.joinpath("config.toml")

    while True:
        with open(config_path, "r") as f:
            results = toml.load(f)

        try:
            config_path = pathlib.Path(results["redirect"])
        except KeyError:
            break

    return results



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import base64
import os
import pathlib
import subprocess
import sys
from datetime import datetime

from pymedphys._imports import imageio
from pymedphys._imports import numpy as np
from pymedphys._imports import pandas as pd
from pymedphys._imports import plt
from pymedphys._imports import streamlit as st
from pymedphys._imports import timeago

import pymedphys
from pymedphys._streamlit.apps.metersetmap import (
    _config,
    _dicom,
    _icom,
    _monaco,
    _mosaiq,
    _trf,
)
from pymedphys._streamlit.utilities import config as st_config
from pymedphys._streamlit.utilities import exceptions as _exceptions
from pymedphys._streamlit.utilities import misc as st_misc

DATA_OPTION_LABELS = {
    "monaco": "Monaco tel.1 filepath",
    "dicom": "DICOM RTPlan file upload",
    "icom": "iCOM record timestamp",
    "trf": "Linac Backup `.trf` filepath",
    "mosaiq": "Mosaiq SQL query",
}

LEAF_PAIR_WIDTHS = (10,) + (5,) * 78 + (10,)
MAX_LEAF_GAP = 410
GRID_RESOLUTION = 1
GRID = pymedphys.metersetmap.grid(
    max_leaf_gap=MAX_LEAF_GAP,
    grid_resolution=GRID_RESOLUTION,
    leaf_pair_widths=LEAF_PAIR_WIDTHS,
)
COORDS = (GRID["jaw"], GRID["mlc"])


def sidebar_overview():
    overview_placeholder = st.sidebar.empty()

    def set_overview_data(patient_id, patient_name, total_mu):
        overview_placeholder.markdown(
            f"Patient ID: `{patient_id}`\n\n"
            f"Patient Name: `{patient_name}`\n\n"
            f"Total MU: `{total_mu}`"
        )

    return set_overview_data


def get_most_recent_file_and_print(linac_id, filepaths):
    if not isinstance(filepaths, list):
        raise ValueError("Filepaths needs to be a list")

    latest_filepath = max(filepaths, key=os.path.getmtime)

    most_recent = datetime.fromtimestamp(os.path.getmtime(latest_filepath))
    now = datetime.now()

    if most_recent > now:
        most_recent = now

    human_readable = timeago.format(most_recent, now)

    st.sidebar.markdown(f"{linac_id}: `{human_readable}`")


def icom_status(linac_id, icom_directory):
    filepaths = list(pathlib.Path(icom_directory).glob("*.txt"))
    get_most_recent_file_and_print(linac_id, filepaths)


def trf_status(linac_id, backup_directory):
    directory = pathlib.Path(backup_directory).joinpath(linac_id)
    filepaths = list(directory.glob("*.zip"))
    get_most_recent_file_and_print(linac_id, filepaths)


def show_status_indicators():
    if st.sidebar.button("Check status of iCOM and backups"):
        try:
            linac_icom_live_stream_directories = (
                _config.get_icom_live_stream_directories()
            )
            linac_indexed_backups_directory = _config.get_indexed_backups_directory()
        except KeyError:
            st.sidebar.write(
                _exceptions.ConfigMissing(
                    "iCOM and/or TRF backup configuration is missing. "
                    "Unable to show status."
                )
            )

            return

        linac_ids = list(linac_icom_live_stream_directories.keys())

        st.sidebar.markdown(
            """
            ## Last recorded iCOM stream
            """
        )

        for linac_id, icom_directory in linac_icom_live_stream_directories.items():
            icom_status(linac_id, icom_directory)

        st.sidebar.markdown(
            """
            ## Last indexed backup
            """
        )

        for linac_id in linac_ids:
            trf_status(linac_id, linac_indexed_backups_directory)


def display_deliveries(deliveries):
    if not deliveries:
        return 0

    st.write(
        """
        #### Overview of selected deliveries
        """
    )

    data = []
    for delivery in deliveries:
        num_control_points = len(delivery.mu)

        if num_control_points != 0:
            total_mu = delivery.mu[-1]
        else:
            total_mu = 0

        data.append([total_mu, num_control_points])

    columns = ["MU", "Number of Data Points"]
    df = pd.DataFrame(data=data, columns=columns)
    st.write(df)

    total_mu = round(df["MU"].sum(), 1)

    st.write(f"Total MU: `{total_mu}`")

    return total_mu


def get_input_data_ui(
    overview_updater_map,
    data_method_map,
    default_method,
    key_namespace,
    advanced_mode_local,
    **previous_results,
):
    if advanced_mode_local:
        data_method_options = list(data_method_map.keys())
        data_method = st.selectbox(
            "Data Input Method",
            data_method_options,
            index=data_method_options.index(default_method),
        )

    else:
        data_method = default_method

    results = data_method_map[data_method](  # type: ignore
        key_namespace=key_namespace,
        advanced_mode_local=advanced_mode_local,
        **previous_results,
    )

    try:
        total_mu = round(display_deliveries(results["deliveries"]), 1)
    except KeyError:
        total_mu = 0

    try:
        patient_id = results["patient_id"]
    except KeyError:
        patient_id = ""

    try:
        patient_name = results["patient_name"]
    except KeyError:
        patient_name = ""

    overview_updater_map[key_namespace](patient_id, patient_name, total_mu)

    return results


@st.cache
def to_tuple(array):
    return tuple(map(tuple, array))


def plot_gamma_hist(gamma, percent, dist):
    valid_gamma = gamma[~np.isnan(gamma)]

    plt.hist(valid_gamma, 50, density=True)
    pass_ratio = np.sum(valid_gamma <= 1) / len(valid_gamma)

    plt.title(
        "Local Gamma ({0}%/{1}mm) | Percent Pass: {2:.2f} % | Mean Gamma: {3:.2f} | Max Gamma: {4:.2f}".format(
            percent, dist, pass_ratio * 100, np.mean(valid_gamma), np.max(valid_gamma)
        )
    )


def plot_and_save_results(
    reference_metersetmap,
    evaluation_metersetmap,
    gamma,
    gamma_options,
    png_record_directory,
    header_text="",
    footer_text="",
):
    reference_filepath = png_record_directory.joinpath("reference.png")
    evaluation_filepath = png_record_directory.joinpath("evaluation.png")
    diff_filepath = png_record_directory.joinpath("diff.png")
    gamma_filepath = png_record_directory.joinpath("gamma.png")

    diff = evaluation_metersetmap - reference_metersetmap

    imageio.imwrite(reference_filepath, reference_metersetmap)
    imageio.imwrite(evaluation_filepath, evaluation_metersetmap)
    imageio.imwrite(diff_filepath, diff)
    imageio.imwrite(gamma_filepath, gamma)

    largest_metersetmap = np.max(
        [np.max(evaluation_metersetmap), np.max(reference_metersetmap)]
    )
    largest_diff = np.max(np.abs(diff))

    widths = [1, 1]
    heights = [0.5, 1, 1, 1, 0.4]
    gs_kw = dict(width_ratios=widths, height_ratios=heights)

    fig, axs = plt.subplots(5, 2, figsize=(10, 16), gridspec_kw=gs_kw)
    gs = axs[0, 0].get_gridspec()

    for ax in axs[0, 0:]:
        ax.remove()

    for ax in axs[1, 0:]:
        ax.remove()

    for ax in axs[4, 0:]:
        ax.remove()

    ax_header = fig.add_subplot(gs[0, :])
    ax_hist = fig.add_subplot(gs[1, :])
    ax_footer = fig.add_subplot(gs[4, :])

    ax_header.axis("off")
    ax_footer.axis("off")

    ax_header.text(0, 0, header_text, ha="left", wrap=True, fontsize=21)
    ax_footer.text(0, 1, footer_text, ha="left", va="top", wrap=True, fontsize=6)

    plt.sca(axs[2, 0])
    pymedphys.metersetmap.display(
        GRID, reference_metersetmap, vmin=0, vmax=largest_metersetmap
    )
    axs[2, 0].set_title("Reference MetersetMap")

    plt.sca(axs[2, 1])
    pymedphys.metersetmap.display(
        GRID, evaluation_metersetmap, vmin=0, vmax=largest_metersetmap
    )
    axs[2, 1].set_title("Evaluation MetersetMap")

    plt.sca(axs[3, 0])
    pymedphys.metersetmap.display(
        GRID, diff, cmap="seismic", vmin=-largest_diff, vmax=largest_diff
    )
    plt.title("Evaluation - Reference")

    plt.sca(axs[3, 1])
    pymedphys.metersetmap.display(GRID, gamma, cmap="coolwarm", vmin=0, vmax=2)
    plt.title(
        "Local Gamma | "
        f"{gamma_options['dose_percent_threshold']}%/"
        f"{gamma_options['distance_mm_threshold']}mm"
    )

    plt.sca(ax_hist)
    plot_gamma_hist(
        gamma,
        gamma_options["dose_percent_threshold"],
        gamma_options["distance_mm_threshold"],
    )

    return fig


@st.cache(hash_funcs={pymedphys.Delivery: hash})
def calculate_metersetmap(delivery):
    return delivery.metersetmap(
        max_leaf_gap=MAX_LEAF_GAP,
        grid_resolution=GRID_RESOLUTION,
        leaf_pair_widths=LEAF_PAIR_WIDTHS,
    )


def calculate_batch_metersetmap(deliveries):
    metersetmap = calculate_metersetmap(deliveries[0])

    for delivery in deliveries[1::]:
        metersetmap = metersetmap + calculate_metersetmap(delivery)

    return metersetmap


@st.cache
def calculate_gamma(reference_metersetmap, evaluation_metersetmap, gamma_options):
    gamma = pymedphys.gamma(
        COORDS,
        to_tuple(reference_metersetmap),
        COORDS,
        to_tuple(evaluation_metersetmap),
        **gamma_options,
    )

    return gamma


def advanced_debugging():
    config = st_config.get_config()

    st.sidebar.markdown("# Advanced Debugging")
    if st.sidebar.button("Compare Baseline to Output Directory"):
        st.write(
            """
            ## Comparing Results to Baseline
            """
        )

        baseline_directory = pathlib.Path(
            config["debug"]["baseline_directory"]
        ).resolve()

        png_baseline_directory = baseline_directory.joinpath("png")

        baseline_png_paths = [
            path for path in (png_baseline_directory.rglob("*")) if path.is_file()
        ]

        relative_png_paths = [
            path.relative_to(png_baseline_directory) for path in baseline_png_paths
        ]

        output_dir = pathlib.Path(config["output"]["png_directory"]).resolve()

        evaluation_png_paths = [
            output_dir.joinpath(path) for path in relative_png_paths
        ]

        for baseline, evaluation in zip(baseline_png_paths, evaluation_png_paths):

            st.write(f"### {baseline.parent.name}/{baseline.name}")

            st.write(f"`{baseline}`\n\n**vs**\n\n`{evaluation}`")

            baseline_image = imageio.imread(baseline)

            try:
                evaluation_image = imageio.imread(evaluation)
            except FileNotFoundError as e:
                st.write(
                    """
                    #### File was not found
                    """
                )

                st.write(e)

                st.write(
                    f"""
                    For debugging purposes, here are all the files that
                    were found within {str(output_dir)}
                    """
                )

                st.write(
                    [str(path) for path in output_dir.rglob("*") if path.is_file()]
                )

                return

            agree = np.allclose(baseline_image, evaluation_image)
            st.write(f"Images Agree: `{agree}`")


def run_calculation(
    reference_results,
    evaluation_results,
    gamma_options,
    escan_directory,
    png_output_directory,
):
    st.write("Calculating Reference MetersetMap...")
    reference_metersetmap = calculate_batch_metersetmap(reference_results["deliveries"])

    st.write("Calculating Evaluation MetersetMap...")
    evaluation_metersetmap = calculate_batch_metersetmap(
        evaluation_results["deliveries"]
    )

    st.write("Calculating Gamma...")
    gamma = calculate_gamma(
        reference_metersetmap, evaluation_metersetmap, gamma_options
    )

    patient_id = reference_results["patient_id"]

    st.write("Creating figure...")
    output_base_filename = (
        f"{patient_id} {reference_results['identifier']} vs "
        f"{evaluation_results['identifier']}"
    )
    pdf_filepath = str(
        escan_directory.joinpath(f"{output_base_filename}.pdf").resolve()
    )
    png_record_directory = png_output_directory.joinpath(output_base_filename)
    png_record_directory.mkdir(exist_ok=True)
    png_filepath = str(png_record_directory.joinpath("report.png").resolve())

    try:
        patient_name_text = f"Patient Name: {reference_results['patient_name']}\n"
    except KeyError:
        patient_name_text = ""

    header_text = (
        f"Patient ID: {patient_id}\n"
        f"{patient_name_text}"
        f"Reference: {reference_results['identifier']}\n"
        f"Evaluation: {evaluation_results['identifier']}\n"
    )

    reference_path_strings = "\n    ".join(
        [str(path.resolve()) for path in reference_results["data_paths"]]
    )
    evaluation_path_strings = "\n    ".join(
        [str(path.resolve()) for path in evaluation_results["data_paths"]]
    )

    footer_text = (
        f"reference path(s): {reference_path_strings}\n"
        f"evaluation path(s): {evaluation_path_strings}\n"
        f"png record: {png_filepath}"
    )

    fig = plot_and_save_results(
        reference_metersetmap,
        evaluation_metersetmap,
        gamma,
        gamma_options,
        png_record_directory,
        header_text=header_text,
        footer_text=footer_text,
    )

    fig.tight_layout()

    st.write("## Results")
    st.pyplot(fig)

    st.write("## Saving reports")
    st.write("### PNG")
    st.write("Saving figure as PNG...")
    plt.savefig(png_filepath, dpi=100)
    st.write(f"Saved:\n\n`{png_filepath}`")
    convert_png_to_pdf(png_filepath, pdf_filepath)


def convert_png_to_pdf(png_filepath, pdf_filepath):
    st.write("### PDF")
    st.write("Converting PNG to PDF...")

    try:
        subprocess.check_call(
            f'magick convert "{png_filepath}" "{pdf_filepath}"', shell=True
        )
        success = True
    except subprocess.CalledProcessError:
        try:
            subprocess.check_call(
                f'convert "{png_filepath}" "{pdf_filepath}"', shell=True
            )
            success = True
        except subprocess.CalledProcessError:
            success = False

    if success:
        st.write(f"Created:\n\n`{pdf_filepath}`")

        with open(pdf_filepath, "rb") as f:
            pdf_contents = f.read()

        pdf_filename = pathlib.Path(pdf_filepath).name

        pdf_b64 = base64.b64encode(pdf_contents).decode()
        href = f"""
            <a href="data:file/zip;base64,{pdf_b64}" download='{pdf_filename}'>
                Click to download {pdf_filename}
            </a>
        """
        st.markdown(href, unsafe_allow_html=True)

    else:
        if sys.platform == "win32":
            url_hash_parameter = "#windows"
        else:
            url_hash_parameter = ""

        download_url = (
            f"https://imagemagick.org/script/download.php{url_hash_parameter}"
        )

        st.write(
            _exceptions.UnableToCreatePDF(
                "Please install Image Magick to create PDF reports "
                f"<{download_url}>."
            )
        )


def main():
    st.write(
        """
        # MetersetMap comparison tool

        Tool to compare the MetersetMap between planned and delivery.
        """
    )

    config = st_config.get_config()

    st.sidebar.markdown(
        """
        # MetersetMap Overview
        """
    )

    st.sidebar.markdown(
        """
        ## Reference
        """
    )

    set_reference_overview = sidebar_overview()

    st.sidebar.markdown(
        """
        ## Evaluation
        """
    )

    set_evaluation_overview = sidebar_overview()

    overview_updater_map = {
        "reference": set_reference_overview,
        "evaluation": set_evaluation_overview,
    }

    st.sidebar.markdown(
        """
        # Status indicators
        """
    )

    show_status_indicators()

    st.sidebar.markdown(
        """
        # Advanced options

        Enable advanced functionality by ticking the below.
        """
    )
    advanced_mode = st.sidebar.checkbox("Run in Advanced Mode")

    gamma_options = _config.get_gamma_options(advanced_mode)

    data_option_functions = {
        "monaco": _monaco.monaco_input_method,
        "dicom": _dicom.dicom_input_method,
        "icom": _icom.icom_input_method,
        "trf": _trf.trf_input_method,
        "mosaiq": _mosaiq.mosaiq_input_method,
    }

    default_reference_id = config["data_methods"]["default_reference"]
    default_evaluation_id = config["data_methods"]["default_evaluation"]
    available_data_methods = config["data_methods"]["available"]

    default_reference = DATA_OPTION_LABELS[default_reference_id]
    default_evaluation = DATA_OPTION_LABELS[default_evaluation_id]

    data_method_map = {}
    for method in available_data_methods:
        data_method_map[DATA_OPTION_LABELS[method]] = data_option_functions[method]

    st.write(
        """
        ## Selection of data to compare
        """
    )

    st.write(
        """
        ### Reference
        """
    )

    reference_results = get_input_data_ui(
        overview_updater_map,
        data_method_map,
        default_reference,
        "reference",
        advanced_mode,
    )

    st.write(
        """
        ### Evaluation
        """
    )

    evaluation_results = get_input_data_ui(
        overview_updater_map,
        data_method_map,
        default_evaluation,
        "evaluation",
        advanced_mode,
        **reference_results,
    )

    st.write(
        """
        ## Output Locations
        """
    )

    st.write(
        """
        ### eSCAN Directory

        The location to save the produced pdf report.
        """
    )

    default_site = evaluation_results.get("site", None)
    if default_site is None:
        default_site = reference_results.get("site", None)

    _, escan_directory = st_misc.get_site_and_directory(
        "eScan Site", "escan", default=default_site, key="escan_export_site_picker"
    )

    escan_directory = pathlib.Path(os.path.expanduser(escan_directory)).resolve()

    if advanced_mode:
        st.write(escan_directory)

    default_png_output_directory = config["output"]["png_directory"]

    if advanced_mode:
        st.write(
            """
            ### Image record

            Path to save the image of the results for posterity
            """
        )

        png_output_directory = pathlib.Path(
            st.text_input("png output directory", default_png_output_directory)
        )
        st.write(png_output_directory.resolve())

    else:
        png_output_directory = pathlib.Path(default_png_output_directory)

    png_output_directory = pathlib.Path(
        os.path.expanduser(png_output_directory)
    ).resolve()

    st.write(
        """
        ## Calculation
        """
    )

    if st.button("Run Calculation"):

        st.write(
            """
            ### MetersetMap usage warning
            """
        )

        st.warning(pymedphys.metersetmap.WARNING_MESSAGE)

        st.write(
            """
            ### Calculation status
            """
        )

        run_calculation(
            reference_results,
            evaluation_results,
            gamma_options,
            escan_directory,
            png_output_directory,
        )

    if advanced_mode:
        advanced_debugging()



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import subprocess
import pytest

import pymedphys

from pymedphys._utilities.test import process

HERE = pathlib.Path(__file__).parent.resolve()


@pytest.mark.cypress
def test_cypress():
    pymedphys.zip_data_paths("metersetmap-gui-e2e-data.zip", extract_directory=HERE)

    pymedphys.zip_data_paths(
        "dummy-ct-and-struct.zip",
        extract_directory=HERE.joinpath("cypress", "fixtures"),
    )

    with process("poetry run pymedphys gui", cwd=HERE, shell=True) as _:
        subprocess.check_call("yarn", cwd=HERE, shell=True)

        subprocess.check_call("yarn cypress run", cwd=HERE, shell=True)



# pylint: disable = pointless-statement, pointless-string-statement
# pylint: disable = no-value-for-parameter, expression-not-assigned
# pylint: disable = too-many-lines, redefined-outer-name

import pathlib

from pymedphys._imports import streamlit as st

from . import exceptions, misc


def monaco_patient_directory_picker(
    patient_id="", key_namespace="", advanced_mode_local=False, site=None
):
    monaco_site, monaco_directory = misc.get_site_and_directory(
        "Monaco Plan Location",
        "monaco",
        default=site,
        key=f"{key_namespace}_monaco_site",
    )

    if advanced_mode_local:
        st.write(monaco_directory.resolve())

    patient_id = st.text_input(
        "Patient ID", patient_id, key=f"{key_namespace}_patient_id"
    )
    if advanced_mode_local:
        patient_id

    if patient_id == "":
        st.stop()

    plan_directories = list(monaco_directory.glob(f"*~{patient_id}/plan"))
    if len(plan_directories) == 0:
        if patient_id != "":
            st.write(
                exceptions.NoRecordsFound(
                    f"No Monaco plan directories found for patient ID {patient_id}"
                )
            )
            st.stop()

        return {"patient_id": patient_id}
    elif len(plan_directories) > 1:
        raise ValueError(
            "More than one patient plan directory found for this ID, "
            "please only have one directory per patient. "
            "Directories found were "
            f"{', '.join([str(path.resolve()) for path in plan_directories])}"
        )

    plan_directory = plan_directories[0]
    patient_directory = pathlib.Path(plan_directory).parent

    return monaco_site, monaco_directory, patient_id, plan_directory, patient_directory



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import pathlib

from pymedphys._imports import numpy as np
from pymedphys._imports import pydicom

HERE = pathlib.Path(__file__).parent
TEMPLATE_DIR = HERE.joinpath("templates")


def load_templates():
    vmat_example = pydicom.read_file(
        str(TEMPLATE_DIR.joinpath("vmat_example.dcm")), force=True
    )

    fff_example = pydicom.read_file(
        str(TEMPLATE_DIR.joinpath("FFF_example.dcm")), force=True
    )

    collimation = pydicom.read_file(
        str(TEMPLATE_DIR.joinpath("24mm_x_20mm_rectangle.dcm")), force=True
    )

    return vmat_example, fff_example, collimation


def from_bipolar(angles):
    ref = angles < 0
    angles[ref] = angles[ref] + 360

    return angles


def main(energy, dose_rate, prepend=""):
    total_mu = "{:.6f}".format(float(dose_rate))
    dose_rate = str(int(dose_rate))
    nominal_energy = str(float("".join(i for i in energy if i in "0123456789.")))
    fff = "fff" in str(energy).lower()

    vmat_example, fff_example, collimation = load_templates()

    fff_fluence_mode = fff_example.BeamSequence[0].PrimaryFluenceModeSequence
    beam_collimation = (
        collimation.BeamSequence[0]
        .ControlPointSequence[0]
        .BeamLimitingDevicePositionSequence
    )

    gantry_step_size = 15.0

    gantry_beam_1 = from_bipolar(np.arange(-180, 181, gantry_step_size))
    gantry_beam_1[0] = 180.1
    gantry_beam_1[-1] = 179.9

    coll_beam_1 = from_bipolar(np.arange(-180, 1, gantry_step_size / 2))
    coll_beam_1[0] = 180.1

    gantry_beam_2 = from_bipolar(np.arange(180, -181, -gantry_step_size))
    gantry_beam_2[-1] = 180.1
    gantry_beam_2[0] = 179.9

    coll_beam_2 = from_bipolar(np.arange(180, -1, -gantry_step_size / 2))
    coll_beam_2[0] = 179.9

    gant_directions = ["CW", "CC"]
    coll_directions = ["CC", "CW"]

    control_point_sequence_beam1 = create_control_point_sequence(
        vmat_example.BeamSequence[0],
        beam_collimation,
        coll_directions[0],
        dose_rate,
        gantry_beam_1,
        coll_beam_1,
        nominal_energy,
    )
    control_point_sequence_beam2 = create_control_point_sequence(
        vmat_example.BeamSequence[1],
        beam_collimation,
        coll_directions[1],
        dose_rate,
        gantry_beam_2,
        coll_beam_2,
        nominal_energy,
    )

    plan = copy.deepcopy(vmat_example)
    plan.BeamSequence[0].ControlPointSequence = control_point_sequence_beam1
    plan.BeamSequence[1].ControlPointSequence = control_point_sequence_beam2

    num_cps = len(gantry_beam_1)

    for beam_sequence, direction in zip(plan.BeamSequence, gant_directions):
        beam_sequence.NumberOfControlPoints = str(num_cps)
        beam_sequence.BeamName = (
            f"WLutzArc-{prepend}-{energy}-{dose_rate.zfill(4)}-{direction}"
        )

    if fff:
        for beam_sequence in plan.BeamSequence:
            beam_sequence.PrimaryFluenceModeSequence = fff_fluence_mode

    plan.FractionGroupSequence[0].ReferencedBeamSequence[0].BeamMeterset = total_mu
    plan.FractionGroupSequence[0].ReferencedBeamSequence[1].BeamMeterset = total_mu

    plan.RTPlanLabel = f"{prepend}-{energy}-{dose_rate}"
    plan.RTPlanName = plan.RTPlanLabel
    plan.PatientID = "WLutzArc"

    return plan


def create_control_point_sequence(
    beam, beam_collimation, rotation_direction, dose_rate, gantry, coll, nominal_energy
):
    num_cps = len(gantry)
    meterset_weight = np.linspace(0, 1, num_cps)

    cp_bounds = bounding_control_points(
        beam, beam_collimation, rotation_direction, dose_rate
    )

    control_point_sequence = pydicom.Sequence([copy.deepcopy(cp_bounds["first"])])

    control_point_sequence[0].GantryAngle = str(gantry[0])
    control_point_sequence[0].BeamLimitingDeviceAngle = str(coll[0])
    control_point_sequence[0].NominalBeamEnergy = nominal_energy

    for i in range(1, num_cps - 1):
        cp = copy.deepcopy(cp_bounds["mid"])

        cp.GantryAngle = str(gantry[i])
        cp.BeamLimitingDeviceAngle = str(coll[i])
        cp.CumulativeMetersetWeight = str(round(meterset_weight[i], 6))
        cp.ControlPointIndex = str(i)

        control_point_sequence.append(cp)

    cp = copy.deepcopy(cp_bounds["last"])
    cp.GantryAngle = str(gantry[-1])
    cp.BeamLimitingDeviceAngle = str(coll[-1])
    cp.ControlPointIndex = str(num_cps - 1)

    control_point_sequence.append(cp)

    return control_point_sequence


def bounding_control_points(beam, beam_collimation, rotation_direction, dose_rate):
    cps = {}

    cps["first"] = beam.ControlPointSequence[0]
    cps["mid"] = beam.ControlPointSequence[1]
    cps["last"] = beam.ControlPointSequence[-1]

    for cp in cps.values():
        cp.BeamLimitingDevicePositionSequence = beam_collimation
        cp.DoseRateSet = dose_rate

    for key in ["first", "mid"]:
        cps[key].BeamLimitingDeviceRotationDirection = rotation_direction

    return cps
